function saved = simulatie(Excelfile,sdir,sheet,stock_solIGR,stat_col)
% Sander Muns, 2019
% Met dit bestand zijn de cijfers in Muns en Werker 
% disclaimer: Er kunnen op geen enkele wijze rechten aan de output van deze code worden
% ontleend.
% Excelfile: geeft input-Excelfile weer, hierop wordt ook output weggeschreven
% sdir:      optionele input om directory van Excelfile aan te geven.
% sheet:     optionele input om sheet aan te geven, leeg betekent alle sheets afgaan.
% stock_solIGR: optionele input om aandelenfractie te overschrijven van SV of IGR.
% stat_col:  optionele input om aan te geven in welke kolom statistieken worden weggeschreven.

% De langetermijn CEQ-vervaningsratio's worden weggeschreven in cellen A59:G59 van tabblad stats.

% % voorbeeldaanroep:
% % basismodel
% % Muns en Werker (2019): Tabel 1, Figuur 1 
% % bijlage: Tabel 3 rij (i)
% simulatie('5050_1%_gam5_leegfonds_minret100.xlsx');
% 
% % minret -50%
% % bijlage: Tabel 2, Figuur 2, Tabel 3 rij (ii)
% simulatie('5050_1%_gam5_leegfonds_minret50.xlsx');
% 
% % gamma = 3
% % bijlage: Tabel 3 rij (iii)
% simulatie('5050_1%_gam3_leegfonds_minret100.xlsx');
% 
% % 60/40-mix
% % bijlage: Tabel 3 rij (iv)
% simulatie('6040_1%_gam5_leegfonds_minret100.xlsx');
% 
% % renterisico
% % bijlage: Tabel 3 rij (v)
% simulatie('5050_renterisico_gam5_leegfonds_minret100.xlsx');
% 
% % optimale beleggingsmix
% % bijlage: Tabel 3 rij (vi)
% simulatie('optmix_1%_gam5_leegfonds_minret100.xlsx');
% 
% % alle vier afwijkende aannames in (ii)-(v) tegelijk
% % bijlage: Tabel 3 rij (vii)
% simulatie('6040_renterisico_gam3_leegfonds_minret50.xlsx');

if ~exist('Excelfile','var') || isempty(Excelfile)
    Excelfile = 'input.xlsx';
end
if ~exist('sdir','var') || isempty(sdir)
    sdir  = '';
end
if ~exist('sheet','var') || isempty(sheet)
    sheet  = [];
end
inputfile    = ['./' sdir '/' Excelfile];
disp(['Reading input from ' inputfile])

e       = actxserver('Excel.Application'); % # open Activex server
ewb     = e.Workbooks.Open([pwd '/' sdir '/' Excelfile]); % # open file (enter full path!)
N_Sheets= ewb.Sheets.Count;
sheets  = cell(1,N_Sheets);
for k=1:N_Sheets
    sheets{k} = ewb.Sheets.Item(k).Name;
end
ewb.Close(false);
e.Quit;
delete(e);
fclose('all');
clear ewb N_Sheets

[parcell,profielcell,sheetcell] = readInput(inputfile,sheet,sheets,false);

for i=1:length(parcell)
    if ~isempty(parcell{i})
    clear x_opt;
    par     = parcell{i};
    profiel = profielcell{i};
    if exist('stock_solIGR','var') && ~isempty(stock_solIGR)
        % overschrijven als expliciet ander solidariteitsvermogen is gegeven
        if ~isnumeric(par.stock_sol)
            par.stock_sol   = stock_solIGR;
            par.bond_sol    = 1-stock_solIGR;
            par.euribor_sol = 0;
            par.hcap_sol    = 0;
        end
        if ~isnumeric(par.stock_IGR)
            par.stock_IGR   = stock_solIGR;
            par.bond_IGR    = 1-stock_solIGR;
            par.euribor_IGR = 0;
            par.hcap_IGR   = 0;
        end
    else
        if all(~isnan(par.stock_sol)) && ischar(par.stock_sol) && ismember(par.stock_sol,sheets)
            [~,~,Ctmp]  = xlsread(inputfile,'stats');
            Irow0       = strcmp(Ctmp(:,1), 'sheetnaam' );
            if (isempty(Irow0) || all(~Irow0)), Irow0 = 2; end
            Irow_stock       = strcmp(Ctmp(:,1), 'stocks_assets ' );
            Irow2       = strcmp(Ctmp(:,1), 'bonds_assets  ' );
            Irow3       = strcmp(Ctmp(:,1), 'euribor_assets' );
            Irow4       = strcmp(Ctmp(:,1), 'hcap_assets' );
            Icol        = strcmp(Ctmp(Irow0 ,:), par.stock_sol );
            par.stock_sol   = cell2mat(Ctmp(Irow_stock,find(Icol,1)));
            par.bond_sol    = cell2mat(Ctmp(Irow2,find(Icol,1)));
            par.euribor_sol = cell2mat(Ctmp(Irow3,find(Icol,1)));
            par.hcap_sol    = cell2mat(Ctmp(Irow4,find(Icol,1)));
            clear Ctmp Irow0 Irow1 Irow2 Irow3 Irow4
        end
    end

    if all(~isnan(par.asset_mix)) && ismember(par.asset_mix,sheets)
        [~,~,Ctmp]       = xlsread(inputfile,'stats');
        Irow0            = strcmp(Ctmp(:,1), 'sheetnaam' );
        if (isempty(Irow0) || all(~Irow0)), Irow0 = 2; end
        Irow_stock       = strcmp(Ctmp(:,1), 'stocks_assets ' );
        Irow_bonds       = strcmp(Ctmp(:,1), 'bonds_assets  ' );
        Icol             = strcmp(Ctmp(Irow0 ,:), par.asset_mix );
        par.asset_mix    = 'lineair';

        perc_stocks      = cell2mat(Ctmp(Irow_stock,find(Icol,1)));
        perc_bonds       = cell2mat(Ctmp(Irow_bonds,find(Icol,1)));
        for tmp = {'stock_first_exposure_age','stock_last_work_age','stock_pension_age','stock_85','stock_max_age'}
            s = par.(tmp{1});
            if isnan(s) || ~isnumeric(s{1})
                par.(tmp{1})       = perc_stocks;
            end
        end
        % lineair profiel obv stockfracties op aantal leeftijden
        v_age    = [par.first_exposure_age   par.pension_age-1   ...
            par.pension_age   min(85,par.max_age)   par.max_age];
        v_stocks = [par.stock_first_exposure_age  par.stock_last_work_age  par.stock_pension_age ...
                        par.stock_85               par.stock_max_age];
        I        = ~isnan(v_stocks);
        num_ages = par.max_age + 1 - par.first_exposure_age;
        profiel.stocks = NaN(num_ages, 1);
        profiel.stocks(1:(max(v_age(I))+1-v_age(1))) = linprofiles( v_age(1), v_age(I), v_stocks(I), 1);
        profiel.bond    = perc_bonds*ones(size(profiel.stocks));
        profiel.euribor = 1 - profiel.stocks - profiel.bond;
        
        clear Ctmp Icol Irow0 Irow1
    end

    disp(par)
    if strcmp(par.max_comp_age_return,'max age')
        par.max_comp_age_return = par.max_age+1-1/par.stepyr;
    end
    if strcmp(par.max_comp_age_return,'pension age')
        par.max_comp_age_return = par.pension_age-1/par.stepyr;
    end
    if par.toed==1 || par.toed==2
        if strcmp(par.toed_renteafdekking,'als optimale beleggingsmix')
            toed_renteafdekking = 1-1/par.gamma;
        else
            if ~isnan(par.toed_renteafdekking)
                toed_renteafdekking = par.toed_renteafdekking;
            else
                toed_renteafdekking = 1;
            end
        end
        if strcmp(par.toed_hcapafdekking,'als optimale beleggingsmix')
            toed_hcapafdekking = 1-1/par.gamma;
        else
            if ~isnan(par.toed_hcapafdekking)
                toed_hcapafdekking = par.toed_hcapafdekking;
            else
                toed_hcapafdekking = 1;
            end
        end
    end
    % bepaal enkele karakteristieken
    num_ages            = par.max_age + 1 - par.first_exposure_age;
    num_work_years      = par.pension_age - par.first_exposure_age;
    num_pension_years   = par.max_age + 1 - par.pension_age;
    num_work_periods    = par.stepyr * num_work_years;
    num_pension_periods = par.stepyr * num_pension_years;
    Iwork               = [true(1,num_work_periods) false(1,num_pension_periods)];
    Ipension            = ~Iwork;
    num_cohorts         = par.nsim_hor*par.stepyr + num_pension_periods;
    
    nloops   = ceil((par.nsim*par.nsim_hor*par.stepyr*num_ages)/(2000*100*65));
    par.nsim = ceil(par.nsim/nloops);
    
    % leeftijd per cohort dat alle pensioenmaanden heeft
    v_birth_year_coh           = (1:par.nsim_hor*par.stepyr+num_pension_periods)/par.stepyr - par.max_age;
    v_first_pension_period_coh = max(1,round(par.stepyr*(v_birth_year_coh + par.pension_age - 1) + 1));
    v_last_pension_period_coh  = min(par.stepyr*par.nsim_hor+1, round(par.stepyr*(v_birth_year_coh + par.max_age ) ));
    v_num_pension_periods_coh  = v_last_pension_period_coh - v_first_pension_period_coh + 1;

    v_first_pension_year_coh   = max(1,v_birth_year_coh + par.pension_age);
    v_last_pension_year_coh    = min(par.nsim_hor+1,v_birth_year_coh + par.max_age+1-1/par.stepyr);
    v_first_pension_age_coh    = min(par.max_age+1-1/par.stepyr, v_first_pension_year_coh - v_birth_year_coh);
    v_last_pension_age_coh     = min(par.max_age+1-1/par.stepyr, v_last_pension_year_coh - v_birth_year_coh);
    
    % pensioengrondslag-profiel
    pension_wage_profile = (profiel.wage(Iwork) - profiel.franchise(Iwork))/par.stepyr;
    if ~isnan(par.franchise_part)
        pension_wage_profile = (profiel.wage(Iwork) * (1 - par.franchise_part))/par.stepyr;
    end
    % nutsmaximalisatie (vooralsnog met simpele bisectie-methode):
    bmaxUtil_ben = strcmp(par.stock_sol, 'NCW CRRA(benefits)')    || strcmp(par.stock_IGR, 'NCW CRRA(benefits)');
    bmaxUtil_ncwben = strcmp(par.stock_sol, 'CRRA(NCW benefits)') || strcmp(par.stock_IGR, 'CRRA(NCW benefits)');
    bmaxUtil_avgben = strcmp(par.stock_sol, 'CRRA(avg.ben)')      || strcmp(par.stock_IGR, 'CRRA(avg.ben)');
    bmaxUtil_sol = strcmp(par.stock_sol, 'NCW CRRA(benefits)')    || strcmp(par.stock_sol, 'CRRA(NCW benefits)') || strcmp(par.stock_sol,'CRRA(avg.ben)');
    bmaxUtil_IGR = strcmp(par.stock_IGR, 'NCW CRRA(benefits)')    || strcmp(par.stock_IGR, 'CRRA(NCW benefits)') || strcmp(par.stock_IGR,'CRRA(avg.ben)');
    if bmaxUtil_ben || bmaxUtil_ncwben || bmaxUtil_avgben
        if bmaxUtil_ben + bmaxUtil_ncwben + bmaxUtil_avgben > 1
            error('cannot optimize portfolio mix SV and IGR differently')
        end
        iter   = 1;
        vutil  = NaN(1,1000); % utility per interatie
        vstocks_iter    = NaN(1,1000);
        % eerste 11 stock exposures van grid
        vstocks_iter(1:11) = 0:.1:1;
    end

    % eenjarige overlevingskansen
    Psurv1 = [1; profiel.Psurv(2:end) ./ profiel.Psurv(1:end-1)];
    if isnan(par.contr_IGR)
        par.contr_IGR = 0;
    end
    contr_sol = 1 - par.contr_ind - par.contr_IGR;
    if contr_sol < 0
        if contr_sol < -1e-8
            error('negatieve premie solidariteitsvermogen')
        else % afrondingsverschil
            contr_sol =0;
        end
    end
    
    Inonneg_ages = [];
    if ~isnan(par.max_comp_age_return)
        Inonneg_ages = ismember(par.first_exposure_age:1/par.stepyr:par.max_age+1-1/par.stepyr, par.first_exposure_age:1/par.stepyr:par.max_comp_age_return);
    end
    
    boptimalLC = any(strcmp(par.asset_mix, {'optimal lifecycle','constant obv geaggr. optimal lifecycle'}));
    bcontinue = true;
    while bcontinue
        if ~isempty(par.seed) && ~isnan(par.seed)
            rng(par.seed); % set random number generator
            rng default
        end        
        avg_benefitcoh_0  = [];
        ncw_benefitcoh_0  = [];
        pensioengroei10_0 = [];
        pensioengroei20_0 = [];
        mCRRA_0           = [];
        mCRRA_report_0    = [];
        repl_rate_benavg_wageavg_0 = [];
        repl_rate_ben1_wageavg_0   = [];
        repl_rate_ben1_wagelast_0  = [];
        indexation_0            = [];
        fundratio_beforeindex_0 = [];
        fundratio_afterindex_0  = [];
        fundratio_0_contr_0     = [];
        fundratio_1_contr_0     = [];
        fraction_IGRbuffer_contr_0 = [];
        fraction_IGRbuffer_sumassets_0 = [];
        assets_sol_0 = [];
        sum_assets_0 = [];
        assets1_0    = [];
        assets_end_0 = [];
        mexposure1_0 = [];
        mexposure2_0 = [];
        
        for loop=1:nloops
        
        if nloops>1
            if loop==1
                disp([num2str(nloops) ': loops = '])
            end
            fprintf([num2str(loop) '-'])
        end

        vmatur = unique([0 1 2 5 20  par.optLC_bondmat  par.maturity_sol  par.maturity_IGR]);
        vmatur = vmatur(~isnan(vmatur));
        spath         = './Scenarios/';
        economy = readKNWset(spath, par, vmatur);

        % aantal simulatiejaren en aantal simulaties
        [nsim,num_sim_prds1] = size(economy.stock_returns);
        % disconteringsmatrix
        I = (economy.RTS_maturities< num_ages);
        discount_mat = economy.RTS(:,I,:).^repmat(-economy.RTS_maturities(I),[par.nsim 1 num_sim_prds1]);
        
        if isnumeric(par.disconto)
            discount_mat_util = repmat( par.disconto.^(0:(1/par.stepyr):num_ages-1/par.stepyr) , [par.nsim 1 num_sim_prds1] );
        else
            discount_mat_util = discount_mat;
        end
        % waarde annuiteit per opbouwperiode, nodig om opbouw om te zetten naar verplichtingen
        % dimensie 1: simulatiejaren, dimensie 2: leeftijd (ivm sterftekans), dimensie 3: simulatienr
        annuityprice = getAnnuityPrices(profiel.Psurv, discount_mat, num_work_periods, 1);
        
        if strcmp(par.proj_return,'discounted RTS')
            % koop annuteit in met uitkeringsstroom die ex-ante voldoet aan de Euler-vergelijking
            % obv cumulatie van bijv (9.13) uit http://web.econ.ku.dk/okocg/VM/VM-general/Kapitler%20til%20bog/Ch9-2016-1.pdf
            % c_{t+1}/c_t = [ par.disconto * (1+r_{t+1}) ]^{1/par.gamma}
            payout_growth_profile = (economy.RTS(:,I,:) * par.disconto ).^repmat(economy.RTS_maturities(I)/par.gamma,[par.nsim 1 num_sim_prds1]);
            annuityprice_disc     = getAnnuityPrices(profiel.Psurv, discount_mat, num_work_periods, payout_growth_profile);
            clear payout_growth_profile;
        end
        
        if any(strcmp(par.asset_mix, {'optimal lifecycle','constant obv geaggr. optimal lifecycle'})) || strcmp(par.contract(1),'4') || par.toed==1 || par.toed==2
            % netto waarde toekomstige pensioenpremies,
            % voor betaling van premie in huidige periode.
            I2              = (economy.RTS_maturities(I) >= 0 & economy.RTS_maturities(I) < par.pension_age-par.first_exposure_age); % -1/par.stepyr
            mat_discount    = cumsum( discount_mat(:, I2, :), 2);
            mat_discount    = mat_discount(:,end:-1:1,:);
        end
        clear discount_mat

        asset_init0      = zeros(nsim, num_ages*par.stepyr);
        if all(~isnan(par.hist_returns)) && ~par.startjong % include historic returns
            for t=1:num_work_periods-1
                mtmp = squeeze( economy.RTS(:,:,1) );
                if strcmp(par.hist_returns,'mean RTS')
                    v = mean(sum((par.contr_rate * pension_wage_profile(1:t)').*( mtmp(:, 1:t).^((t:-1:1)/par.stepyr)),2));
                else
                    v = sum((par.contr_rate * pension_wage_profile(1:t)).*(1+par.hist_returns).^((t:-1:1)'/par.stepyr));
                end
                asset_init0(:,t) = repmat(v,[nsim 1]);
            end
            asset_init0(:,num_work_periods:end) = repmat(linspace(v,0,num_pension_periods+1),[nsim 1]);
            v_num_work_periods_coh = repmat(num_work_periods,1,num_cohorts);
            clear mtmp v;
        else
            v_num_work_periods_coh = max(0,min(num_work_periods, v_last_pension_period_coh - v_num_pension_periods_coh - 1));
        end
        
        for s = {'bond_sol','hcap_sol','euribor_sol'}
            if isempty(par.(s{1})) || isnan(par.(s{1}))
                par.(s{1})    = 0;
            end
        end
        if bmaxUtil_sol
            par.stock_sol   = vstocks_iter(iter);
            if strcmp(par.optLC_bondmat,'euribor')
                par.euribor_sol = 1 - par.stock_sol;
            else
                if strcmp(par.optLC_bondmat,'hcap')
                    par.hcap_sol    = 1 - par.stock_sol;
                else
                    par.bond_sol = 1 - par.stock_sol;
                end
            end
        end
        if bmaxUtil_IGR
            par.bond_IGR    = 0;
            par.hcap_IGR    = 0;
            par.euribor_IGR = 0;
            par.stock_IGR   = vstocks_iter(iter);
            if strcmp(par.optLC_bondmat,'euribor')
                par.euribor_IGR = 1 - par.stock_IGR;
            else
                if strcmp(par.optLC_bondmat,'hcap')
                    par.hcap_IGR    = 1 - par.stock_IGR;
                else
                    par.bond_IGR = 1 - par.stock_IGR;
                end
            end
        end
        % initialiseer:
        % uitkeringen
        benefit_periods = zeros(nsim,num_pension_periods,num_sim_prds1-1);
        % initialiseer premie-matrix
        contr           = zeros(nsim, num_ages*par.stepyr, num_sim_prds1-1);
        
        benefityr       = NaN(nsim, num_pension_years, par.nsim_hor);
        if strcmp(par.contract(1),'1') || strcmp(par.contract,'4B')
            % opbouwprofiel
            accr_profile = 1 ./ annuityprice(:, Iwork, :); % degressieve opbouw
            % opbouw
            accrual      = zeros(nsim, num_ages*par.stepyr, num_sim_prds1);
            if ~isnan(par.hist_returns) % historische opbouw toevoegen
                accrual(:,Iwork,1) = contr_sol .* par.contr_rate .* cumsum( repmat(mean(accr_profile(:,:,1)) ...
                    .* pension_wage_profile', [nsim 1]), 2); % .* squeeze(economy.RTS(:, 1:t, 1)).^(t:-1:1)'
                accrual(:,Ipension,1) = repmat(accrual(:,num_work_years*par.stepyr,1), [1 num_pension_periods]);
            end
            % verplichtingen
            accrualPVsum    = zeros(nsim, num_sim_prds1);
            % uitbetalingen
            benefit_periods(:,:,1)  = accrual(:, Ipension, 1);
            benefitsum      = zeros(nsim, num_sim_prds1);
            benefitsum(:,1) = squeeze(sum(benefit_periods(:,:,1),2))';
            % verwaarloos indexatie bij initialisatie
            accrualPVsum(:,1) = squeeze( sum( annuityprice(:,:,1) .* accrual(:,:,1), 2) ) - benefitsum(:,1);
            accrualPVsum_end  = accrualPVsum;
            if strcmp(par.contract(1),'1') && ~par.startjong && any(~isnan(par.hist_returns)) && ~isnan(par.init_dkg)
                init_scaling_factor = par.init_dkg ./ (squeeze( sum(asset_init0,2) ) ./ accrualPVsum(:,1));
                asset_init0 = init_scaling_factor .* asset_init0;
            end
            % dekkingsgraad
            fundratio_beforeindex = repmat(par.init_dkg, [nsim num_sim_prds1]);
            fundratio_afterindex  = repmat(par.init_dkg, [nsim num_sim_prds1]);
            fundratio_0_contr     = NaN(nsim,num_sim_prds1);
            fundratio_1_contr     = NaN(nsim,num_sim_prds1);
            % indexatiematrix, allemaal 1 betekent geen vaste indexatie maar evt
            % wel obv dekkingsgraad.
            indexation0 = ones(nsim, 1, num_sim_prds1);
            % indexatiematrix
            indexation  = ones(nsim, num_sim_prds1);
            if strcmp(par.contract,'1A')
                % kortingsjaren
                cutyears = ones(nsim, num_sim_prds1);
                % onconditioneel korten
                unc_cut1A = zeros(nsim, num_sim_prds1);
            end
        else
            accrualPVsum            = [];
            accrualPVsum_end        = [];
            fundratio_beforeindex = [];
            fundratio_afterindex  = [];
            indexation            = [];
            fundratio_0_contr         = [];
            fundratio_1_contr         = [];
        end
        
        if strcmp(par.contract(1),'1')
            assets      = [];   % geen persoonlijk pensioenvermogen
            assets_IGR  = [];   % geen IGR-vermogen
        else
            % PPV
            assets             = zeros(nsim, num_ages * par.stepyr, num_sim_prds1-1);
            assets(:,:,1)      = asset_init0;
            % IGR
            if par.contr_IGR~=0
                assets_IGR  = zeros(nsim, num_sim_prds1-1);
                ret_IGR = zeros(nsim, num_sim_prds1-1);
            else
                assets_IGR  = [];   % geen IGR-vermogen
                ret_IGR = [];
            end
        end
        if strcmp(par.contract(1),'4') 
            if (boptimalLC || par.toed==1 || par.toed==2) && strcmp(par.set_RTS,'Vasicek') 
                vhorizon_pension   = 1/par.stepyr : 1/par.stepyr : (par.max_age+1-1/par.stepyr - par.first_exposure_age);
                v_wT_pension       = zeros(size(vhorizon_pension));
                % toekomstige jaren waarin human capital kan worden opgebouwd, ongeacht de leeftijd
                vhorizon_hcap      = 1/par.stepyr : 1/par.stepyr : (par.pension_age-1/par.stepyr - par.first_exposure_age);
                v_wT_hcap_age      = zeros(length(vhorizon_hcap), num_work_periods);
                for age=par.first_work_age : 1/par.stepyr : par.max_age+1-1/par.stepyr
                    vhorizon_pension_age           = (par.pension_age : 1/par.stepyr: par.max_age+1-1/par.stepyr) - age;
                    vhorizon_pension_age(vhorizon_pension_age<=0) = [];
                    Ihorizon_age                   = ismember(vhorizon_pension, vhorizon_pension_age);
                    v_wT_pension(Ihorizon_age)     = v_wT_pension(Ihorizon_age) + 1;
                    if age < par.pension_age-1/par.stepyr % alleen dan human capital
                        % horizon toekomstige jaren waarin human capital wordt opgebouwd door iemand van leeftijd age
                        vhorizon_hcap_age        = (1/par.stepyr : 1/par.stepyr : par.pension_age-1/par.stepyr-age );
                        vhorizon_hcap_age(vhorizon_hcap_age<=0) = [];
                        Ihorizon_age             = ismember(vhorizon_hcap, vhorizon_hcap_age);
                        futurewages              = pension_wage_profile(age+1+1-par.first_exposure_age:end);
                        v_wT_hcap_age(Ihorizon_age, age+1-par.first_exposure_age)  = flipud( futurewages );
                    end
                end
                v_wT_pension = v_wT_pension(:)/sum(v_wT_pension);
                B_vhorizon_pension = B_BandX(vhorizon_pension, par.V_alp, 1/par.stepyr);
                
                % exposure op basis van assets, ook per scenario andere exposure
                if isequal(par.optLC_bondmat,'matching')
                    % omdat er n risicofactor is, zijn exposures van verschillende looptijden lineaire combinaties
                    % er wordt daarom in obligaties van n looptijd genvesteerd
                    B_optLC_bondmat = B_vhorizon_pension * v_wT_pension;
                else
                    B_optLC_bondmat = B_BandX(par.optLC_bondmat, par.V_alp, 1/par.stepyr);
                end
                % xv2_hcap gemid factorgevoeligheid hcap, corrigeert allociatie voor reeds aanwezig human capital
                xv2_hcap        = zeros( size(Iwork) );
                xv2_hcap(Iwork) = ( B_BandX(vhorizon_hcap, par.V_alp, 1/par.stepyr) * v_wT_hcap_age / B_optLC_bondmat );
                
                piS    =  par.BS_risk_par * par.BS_vol_par;
                piP1   = -par.V_lam * par.V_rho * B_optLC_bondmat;
                sigSr  = par.BS_V_corr * par.V_rho * par.BS_vol_par ;	% tussen (10) en (11) in BX(2000)
                sigP1S = -sigSr * B_optLC_bondmat;
                sigP1  =  par.V_rho * B_optLC_bondmat;
                Sigma  = [par.BS_vol_par^2  sigP1S;
                          sigP1S(:)        sigP1(:)*sigP1];
                vpi    = [piS; piP1(:)];

                sigSP  =  sigSr * B_optLC_bondmat;
                sigP1P =  par.V_rho^2 * B_optLC_bondmat' * B_optLC_bondmat;
                x_opt  = 1/par.gamma * (Sigma\vpi) + (1-1/par.gamma) * (Sigma\ [sigSP; sigP1P]);
                if sigP1==0
                   x_opt(2,:) = 1 - x_opt(1,:); 
                end
                x_opt      = repmat( permute(x_opt, [3 2 1]), [nsim 1 1]);
            end
        end
        if strcmp(par.contract,'4A')
            % geen solidariteitsvermogen
            assets_sol = [];
        else
            % initialiseer solidariteitsvermogen
            assets_sol  = zeros(nsim, par.nsim_hor+1);
            if any(~isnan(par.hist_returns)) && strcmp(par.contract(1),'1')
                assets_sol(:,1)  = squeeze( sum(asset_init0,2) );
            end
            % rendement solidariteitsvermogen
            ret_sol     = zeros(nsim, num_sim_prds1);
        end
        if i > 1 || ~exist('stat_col','var') || isempty(stat_col)
            stat_col = i+1;
        end
        economy = rmfield(economy,'RTS'); % bespaar geheugen
        % zelfde stappen als ALM-model CPB, zie Michielsen (2015)
        % https://www.cpb.nl/sites/default/files/publicaties/download/cpb-background-document-september2015-cpb-asset-liability-management-model-pension-analyses.pdf
        % 1 ontvang premies
        % 2 betaal uitkeringen
        % 3 verwerk financiele rendementen
        % 4 bepaal korting en indexatie

        % eerste dimensie = scenario's
        % tweede dimensie = leeftijd
        % derde dimensie  = simulatiejaren
        % diagonalen      = cohorten

        for t1=2:par.nsim_hor+1
            for t2 = 0:(par.stepyr-1)
                t = 2 + (t1-2)*par.stepyr + t2;
                
                % pensioengrondslag loonprofiel
                pension_wage_profile = (profiel.wage(Iwork) - profiel.franchise(Iwork))/par.stepyr;

                % pensioengrondslag loonprofiel startend met startsalaris voor evt ongeborenen
                pension_wage_profile0   = pension_wage_profile;
                if par.first_work_age > par.first_exposure_age
                    min_work_age            = par.first_work_age - par.first_exposure_age + 1; % aanname: evt ongeborenen rekenen met startsalaris
                    pension_wage_profile0(1:(min_work_age-1)) = pension_wage_profile(min_work_age);
                end
                % bepaal rendement op human capital
                Iworkt = Iwork;
                if par.startjong % nu is toekomstig loon niet meer nodig
                    pension_wage_profile(t:end) = 0;
                    Iworkt(t:end) = false;
                end
                if any(strcmp(par.asset_mix, {'optimal lifecycle','constant obv geaggr. optimal lifecycle'})) || par.toed==1 || par.toed==2 || strcmp(par.payout_meth_IGR, "bereikbaar pensioen")
                    % bepaal rendement op human capital, -1 aftrekken omdat vorige premieinleg niet meer meetelt:
                    hcap_ret_t                 = ones(nsim, num_ages);
                    hcap_ret_t(:,1:num_work_periods-1) = repmat(pension_wage_profile0(2:end)'  , [nsim 1]) .* squeeze( mat_discount(:, 2:end  , t+1 ) ) ...
                                                    ./ ( repmat(pension_wage_profile0(1:end-1)', [nsim 1]) .* squeeze( mat_discount(:, 1:end-1, t   ) - 1 ) );
                    hcap_ret_t(isnan(hcap_ret_t) | isinf(hcap_ret_t)) = 1;
                    
                    hcap            = par.contr_rate .* repmat(pension_wage_profile0', [nsim 1]) .* squeeze( mat_discount(:,:,t) - 1 );
                    hcap(:,~Iworkt & Iwork) = 0;
                else
                    hcap_ret_t = 1;
                end

                % bepaal pensioenpremie
                contr(:,Iwork,t)           = repmat( par.contr_rate .* pension_wage_profile' , [nsim 1]);
                if strcmp(par.contract(1),'4')
                    % begin nieuwe periode met individuele vermogen uit vorige periode:
                    assets(:,2:end,t) = assets(:,1:end-1,t-1); % .* repmat( 1-Pdie(1:end-1), [1 1 nsim]); SMU: door biometrisch rendement hier geen aanpassing
                    % voeg pensioenpremie toe
                    assets(:,:,t)     = assets(:,:,t) + par.contr_ind .* contr(:,:,t);
                    % pas IGR-buffer aan als deze bestaat
                    if par.contr_IGR ~= 0
                        % verwerk premie en uitbetalingen aan IGR-buffer
                        assets_IGR(:,t)   = assets_IGR(:,t-1);
                        assets_IGR(:,t)   = assets_IGR(:,t) + sum( par.contr_IGR .* contr(:,:,t), 2);
                        if (par.payoutratio_sol~=0 && isnan(par.payoutratio_sol))
                            if strcmp(par.payout_meth_IGR, "inlegprofiel")
                                payoutprofile_IGR = par.contr_rate .* pension_wage_profile(Iwork);
                                payoutprofile_IGR = payoutprofile_IGR ./ sum( payoutprofile_IGR );
                                assets(:,Iwork,t) = assets(:,Iwork,t) + par.payoutratio_IGR / par.stepyr * assets_IGR(:,t) .* payoutprofile_IGR(Iwork)';
                            else
                                if strcmp(par.payout_meth_IGR, "bereikbaar pensioen")                                    
                                    % total wealth = financial wealth + human capital
                                    total_wealth_t = NaN(nsim, num_work_periods);
                                    % werkenden
                                    total_wealth_t(:,1:num_work_periods) = assets(:,1:num_work_periods,t) + hcap;
                                    payoutprofile_IGR = total_wealth_t ./ sum( total_wealth_t, 2);
                                    % IGR-premie wordt per definitie eerst teruggegeven, dit wordt verrekend
                                    payout_amount     = par.payoutratio_IGR / par.stepyr * assets_IGR(:,t) - par.contr_IGR .* sum(contr(:,Iwork,t), 2);
                                    assets(:,Iwork,t) = assets(:,Iwork,t) + par.contr_IGR .* contr(:,Iwork,t) + payout_amount .* payoutprofile_IGR(:,Iwork);
                                end
                            end
                            if isnan(par.payoutratio_IGR) && (par.contr_IGR~=0 && ~isnan(par.contr_IGR))
                                warning('no IGR payoutratio => ratio set at 100%')
                                par.payoutratio_IGR = 1;
                            end
                        end
                        assets_IGR(:,t)   = (1-par.payoutratio_IGR/par.stepyr) * assets_IGR(:,t);
                    end
                    if strcmp(par.contract,'4A')
                        % bepaal uitkering als WVP met annuiteitsfactor van projectierente
                        switch par.proj_return
                            case 'RTS'
                                benefit_t       = assets(:,Ipension,t) ./ annuityprice(:,Ipension,t) ;
                            case 'discounted RTS' % eerste uitkeringen zijn belangrijker
                                benefit_t       = assets(:,Ipension,t) ./ annuityprice_disc(:,Ipension,t) ;
                        end
                        if par.smoothper~=1
                            benefit_periods(:, 1, t) = benefit_periods(:, 1, t) + benefit_t(:,1);
                            % uitsmeren als er vorige periode al een uitkering was
                            benefit_periods(:, 2:end, t) = benefit_periods(:, 2:end, t) + benefit_periods(:,1:end-1,t-1) + (benefit_t(:,2:end) - benefit_periods(:,1:end-1,t-1)) / par.smoothper;
                        else
                            benefit_periods(:,:, t) = benefit_periods(:,:,t) + benefit_t;
                        end
                        % nooit meer uitbetalen dan aanwezig in persoonlijk pensioenvermogen;
                        benefit_periods(:,:,t)       = min( benefit_t, assets(:,Ipension,t) );
                        % betaal uitkering uit individuele vermogens
                        assets(:, Ipension,t) = assets(:,Ipension,t) - benefit_periods(:,:,t);
                    end
                end
                if strcmp(par.contract(1),'1') || strcmp(par.contract,'4B')
                    % begin nieuwe periode met sol. vermogen en opbouw uit vorige periode:
                    assets_sol(:,t)      = assets_sol(:,t-1);
                    % voeg premie toe aan solidariteitsvermogen
                    assets_sol(:,t)      = assets_sol(:,t) + sum( contr_sol .* contr(:,:,t), 2) ;

                    accrual(:, 2:end, t) = accrual(:,1:end-1,t-1);
                    % voeg nieuwe opbouw toe
                    vtmp = 1;
                    if ~isnan(par.adjaccrfactor) && par.adjaccrfactor~=0
                        % evt dekkingsgraadneutraal inkopen
                        vtmp = (1 + (1./fundratio_afterindex(:,t-1)-1) * par.adjaccrfactor);
                        vtmp(isnan(vtmp)) = 1;
                    end
                    if strcmp(par.contract(1),'1')
                        accrual(:, Iworkt, t) = accrual(:, Iworkt, t) + vtmp .* contr_sol .* par.contr_rate .* accr_profile(:,Iworkt,t) .* repmat(pension_wage_profile(Iworkt)', [nsim 1]);
                    end
                    if strcmp(par.contract,'4B')
                        accrual(:,Iworkt,t)   = accrual(:, Iworkt, t) + vtmp .* repmat(profiel.annuity_buy(Iworkt)', [nsim 1]) .* accr_profile(:,Iworkt,t) .* assets(:,Iworkt,t);
                        % verschuif vermogen van individueel naar sol.vermogen
                        % deze stap is voordat rendementen worden toegekend zodat 1B = 4B bij volledige omzetting vanaf startleeftijd
                        sold_ind_assets      = repmat(profiel.annuity_buy(Iworkt)', [nsim 1]) .* assets(:,Iworkt,t);
                        assets_sol(:,t)      = assets_sol(:,t) + squeeze( sum( sold_ind_assets , 2) );
                        assets(:, Iworkt, t)  = assets(:, Iworkt, t) - sold_ind_assets;
                    end
                    clear vtmp sold_ind_assets

                    % bepaal uitkering aan de hand van opbouw en evt AOW
                    benefit_t               = accrual(:, Ipension, t);
                    benefit_periods(:,:,t)  = benefit_periods(:,:,t) + benefit_t;
                    benefitsum(:,t)         = squeeze(sum(benefit_t,2))';
                    % betaal uitkering uit solidariteitsvermogen
                    assets_sol(:,t)         = assets_sol(:,t) - benefitsum(:,t);

                    % ken beleggingsrendementen toe aan solidariteitsvermogen
                    % bepaal eerst het rendement op de obligatieportefeuille
                    if strcmp(par.maturity_sol,"matching")
                       bond_ret = sum( annuityprice(:,2:end,t+1) .* accrual(:,1:end-1,t) , 2) ./ ...
                           sum( ( annuityprice(:,1:end-1,t) - repmat( (profiel.Psurv(1:end-1) .* ...
                           Ipension(1:end-1)')', [nsim 1]) ) .* accrual(:,1:end-1,t), 2);
                       I = isnan(bond_ret);
                       if any(I)
                           if isnumeric(par.set_RTS)
                               bond_ret(I) = 1 + par.set_RTS;
                           else
                               if isnumeric(par.V_r0)
                                    bond_ret(I) = 1 + par.V_r0;
                               else
                                    bond_ret(I) = 1;
                               end
                           end
                       end
                       if strcmp(par.portfolio_rebalance,'continu')
                           bond_ret = exp(bond_ret-1);
                       end
                    else
                       bond_ret = squeeze( economy.bond_returns(:,economy.bond_portf_maturities==par.maturity_sol,t) );
                    end
                    
                    if strcmp(par.portfolio_rebalance,'continu')
                        ret_sol(:,t) = economy.stock_returns(:,t) .^ par.stock_sol .* ...
                            bond_ret .^ par.bond_sol .* economy.euribor_return(:,t) .^ par.euribor_sol .* prod(hcap_ret_t .^ par.hcap_sol, 2);
                    else
                        ret_sol(:,t) = economy.stock_returns(:,t) .* par.stock_sol + ...
                            bond_ret .* par.bond_sol + economy.euribor_return(:,t) .* par.euribor_sol + sum(hcap_ret_t .* par.hcap_sol, 2);
                    end
                    % ken rendement toe aan solidariteitsvermogen
                    assets_sol(:,t) = ret_sol(:,t) .* assets_sol(:,t);

                    % als geen biometrisch rendement: pas verplichtingen aan op sterfte
                    % pas evt vaste indexatieregel toe op vorige opbouw:
                    accrual(:, 2:end, t) = accrual(:, 2:end, t) .* repmat( Psurv1(1:end-1)', [nsim 1]) ...
                                            .* repmat( indexation0(:,:,t), [1 num_ages*par.stepyr-1]);                    
                    % indexeer/kort alleen aan het begin van de periode:
                    % bereken totale verplichtingen per scenario (bij inleg geldt:
                    % accrualPVsum(tegen einde periode waarde) = assets_sol*(1 + looptijd gewogen risicovrije rente)
                    accrualPVsum(:,t)  = squeeze( sum( annuityprice(:,2:end,t+1) .* accrual(:,1:end-1,t), 2) );
                    % bepaal dekkingsgraad na rendement en vr indexatie
                    if any(accrualPVsum(:,t))~=0
                        fundratio_beforeindex(:,t)  = assets_sol(:,t) ./ accrualPVsum(:,t);
                    end
                    if strcmp(par.contract,'1A')
                        % bepaal evt uitgesmeerde onvoorwaardelijke korting
                        unc_cut1A(cutyears(:,t-1)==0,t )= min(0, fundratio_beforeindex(cutyears(:,t-1)==0,t)-par.unc_cutratio) / par.smoothper;
                        cutyears(cutyears(:,t-1)>0, t)  = cutyears( cutyears(:,t-1)>0, t-1) + 1;
                        cutyears(cutyears(:,t)==0,  t)  = (unc_cut1A(cutyears(:,t)==0,t) < 0);
                        cutyears(cutyears(:,t)==10, t)  = 0;

                        % bepaal indexatie/korting obv prijsindexatie
                        if length(economy.price_inflation)==1
                            infl = economy.price_inflation;
                        else
                            infl = economy.price_inflation(:,t-1);
                        end
                        indexation(:,t) = 1 + max(0, infl-1) ...
                            .* max(0, min( any(accrualPVsum(:,t))*par.ftk_index_high, fundratio_beforeindex(:,t))-par.ftk_index_low)/(par.ftk_index_high - par.ftk_index_low) ...
                            + min(0, fundratio_beforeindex(:,t)-par.ftk_cut) / par.smoothper + unc_cut1A(:,t);
                    elseif strcmp(par.contract,'1B') || strcmp(par.contract,'4B')
                        % bepaal indexatie/korting
                        indexation(:,t) = 1 + (fundratio_beforeindex(:,t)-par.target_dkg) / par.smoothper;
                    end
                    Imin                = (fundratio_beforeindex(:,t) < par.min_covratio);
                    % bepaal indexatie/korting
                    indexation(Imin,t)  = min( indexation(Imin,t), 1 + (fundratio_beforeindex(Imin,t)-par.min_covratio) / par.min_covratio );
                    Imax                = (fundratio_beforeindex(:,t) > par.max_covratio);
                    % bepaal indexatie/korting
                    indexation(Imax,t)  = max( indexation(Imax,t), 1 + (fundratio_beforeindex(Imax,t)-par.max_covratio) / par.max_covratio );
                    indexation( isnan(indexation(:,t)) ,t) = 1;
                    % bepaal opbouw na indexatie
                    accrual0                    = squeeze( accrual(:,:,t) );
                    accrual(:,:,t)              = accrual0 .* indexation(:,t);
                    accrualPVsum_end(:,t)       = squeeze(sum( annuityprice(:,2:end,t+1) .* accrual(:,1:end-1,t),2));
                    fundratio_afterindex(:,t)   = assets_sol(:,t) ./ sum( annuityprice(:,2:end,t+1) .* accrual(:,1:end-1,t),2);
                    fundratio_afterindex( isnan(fundratio_afterindex(:,t)) ,t) = 1;
                    if par.age_index
                        daccrual0               = sum( annuityprice(:,2:end,t+1) .* accrual0(:,1:end-1) .* ( indexation(:,t) - 1), 2);
                        % corrigeer mate van indexatie voor human
                        % capital: [V(L) + H(L)]/V(L) =approx= 1 + (jaren tot pensioen geteld vanaf volgende periode))/werkjaren
                        dindexation1            = (indexation(:,t)-1) * (1 + fliplr(cumsum(fliplr(Iwork(2:end))))./cumsum(Iwork(1:end-1)));
                        daccrual1               = sum( annuityprice(:,2:end,t+1) .* accrual0(:,1:end-1) .* dindexation1, 2);
                        scalefactor             = daccrual0 ./ daccrual1;
                        accrual(:,1:end-1,t)          = accrual0(:,1:end-1)  + scalefactor .* accrual0(:,1:end-1) .* dindexation1;
                        clear fundratio_afterindex0 scalefactor
                    end
                    clear accrual0
                end
                mexposure = zeros(0,0,2);
                if strcmp(par.contract(1),'4')
                    if boptimalLC || par.toed==1 || par.toed==2
                        % total wealth = financial wealth + human capital
                        total_wealth_t = NaN(nsim, num_ages*par.stepyr);
                        % werkenden
                        total_wealth_t(:,1:num_work_periods) = assets(:,1:num_work_periods,t) + hcap;
                        % gepensioneerden
                        total_wealth_t(:,num_work_periods+1:end) = assets(:,num_work_periods+1:end,t);
                        mexposure        = zeros([size(total_wealth_t) 2]);
                        if ~isequal(par.asset_mix,'lineair')
                            % steeds ander profiel op basis van assets, ook per scenario ander profiel
                            if strcmp(par.set_RTS,'Vasicek')
                                mexposure(:,:,1) = min(par.max_exposure, x_opt(:,:,1) .* total_wealth_t ./ assets(:,:,t) );
                                mexposure(:,:,2) = min(par.max_exposure,(x_opt(:,:,2) .* total_wealth_t - xv2_hcap) ./ assets(:,:,t) );
                                if ~isnan(par.max_exposure)
                                    mexposure = min(par.max_exposure, mexposure);
                                end
                            else % geen renterisico: exposures tov total_wealth tellen op tot n
                                c_tmp            = par.BS_risk_par / (par.BS_vol_par * par.gamma);
                                mexposure(:,:,1) = min(par.max_exposure, c_tmp .* total_wealth_t ./ assets(:,:,t));
                                mexposure(:,:,2) = min(par.max_exposure, (1 - c_tmp) .* total_wealth_t ./ assets(:,:,t));
                            end
                            mexposure(repmat(assets(:,:,t)==0,[1 1 2])) = 0; % voorkom NaN als geen assets
                        
                            if isnumeric(par.asset_mix)
                                profiel.stocks = repmat( par.asset_mix   , [num_ages*par.stepyr 1]) ;
                                profiel.bond   = repmat( 1-par.asset_mix , [num_ages*par.stepyr 1]) ;
                            else
                                if strcmp(par.asset_mix,'constant obv geaggr. optimal lifecycle')
                                    profiel.stocks = squeeze( repmat( sum( mexposure(:,:,1) .* assets(:,:,t) , 2) ./ sum( assets(:,:,t), 2), [1 num_ages*par.stepyr]) );
                                    profiel.bond   = squeeze( repmat( sum( mexposure(:,:,2) .* assets(:,:,t) , 2) ./ sum( assets(:,:,t), 2), [1 num_ages*par.stepyr]) );
                                else
                                    profiel.stocks = squeeze( mexposure(:,:,1) );
                                    profiel.bond   = squeeze( mexposure(:,:,2) );
                                end
                            end
                        end
                        if size(profiel.stocks,1)==1
                             profiel.stocks =  profiel.stocks';
                        end
                        profiel.stocks = rescale(profiel.stocks, assets(:,:,t), Inonneg_ages);
                        if strcmp(par.optLC_bondmat,'euribor')
                            profiel.bond = zeros(size(profiel.stocks));
                        else
                            profiel.bond = rescale(profiel.bond, assets(:,:,t), Inonneg_ages);
                        end
                        if isequal(par.asset_mix,'lineair')
                            mexposure(:,:,1) = repmat(profiel.stocks', par.nsim, 1);
                            mexposure(:,:,2) = repmat(profiel.bond', par.nsim ,1);
                        end
                        profiel.euribor = 0;
                        profiel.hcap    = 0;
                        if strcmp(par.short_rate,'human capital')
                            profiel.hcap    = 1 - profiel.stocks - profiel.bond;
                        else
                            profiel.euribor = 1 - profiel.stocks - profiel.bond;
                        end
                    else
                        profiel.hcap    = 0;
                    end
                    if par.toed==1 || par.toed==2 || strcmp(par.optLC_bondmat,'matching')
                        annuityreturn_t                            = ones(nsim,num_ages*par.stepyr);
                        annuityreturn_t(:,1:num_ages*par.stepyr-1) = annuityprice(:,2:end,t+1) ./ ...
                            ( annuityprice(:,1:end-1,t) - repmat( (profiel.Psurv(1:end-1) .* Ipension(1:end-1)')', [nsim 1]) );
                    end
                    % bepaal feitelijk rendement
                    if isnumeric(par.optLC_bondmat) && ~isnan(par.optLC_bondmat)
                        bondret = squeeze( economy.bond_returns(:,economy.bond_portf_maturities==par.optLC_bondmat,t) );
                    else
                        if strcmp(par.optLC_bondmat,'matching')
                            bondret     = annuityreturn_t;
                        else
                            bondret     = ((par.stepyr-t2)/par.stepyr * economy.euribor_return(:,t)  + (t2/par.stepyr) * economy.euribor_return(:,t+1)).^(1/par.stepyr);
                        end
                    end
                    if strcmp(par.portfolio_rebalance,'continu')
                        if boptimalLC 
                            ret_nonstock = bondret .^ profiel.bond .* repmat( economy.euribor_return(:,t), [1 num_ages*par.stepyr]) .^ profiel.euribor ...
                                            .* hcap_ret_t .^ profiel.hcap;
                            ret_t        = repmat( economy.stock_returns(:,t), [1 num_ages*par.stepyr]) .^ profiel.stocks .* ret_nonstock;
                        else
                            ret_nonstock = bondret .^ repmat( profiel.bond(:)', [par.nsim 1]) ...
                                .* repmat( economy.euribor_return(:,t), [1 num_ages*par.stepyr]) .^ repmat( profiel.euribor(:)', [par.nsim 1]) ...
                                .* hcap_ret_t .^repmat( profiel.hcap(:)', [par.nsim 1]);
                            ret_t        = repmat( economy.stock_returns(:,t), [1 num_ages*par.stepyr]) .^ repmat( profiel.stocks(:)', [par.nsim 1]) .* ret_nonstock;
                        end
                    else
                        if boptimalLC
                            ret_nonstock = bondret .* profiel.bond + ... 
                                repmat( economy.euribor_return(:,t), [1 num_ages*par.stepyr]) .* profiel.euribor + hcap_ret_t .* profiel.hcap;
                            ret_t        = repmat( economy.stock_returns(:,t), [1 num_ages*par.stepyr]) .* profiel.stocks + ret_nonstock;
                        else
                            ret_nonstock = bondret .* profiel.bond(:)' + economy.euribor_return(:,t) * profiel.euribor(:)' + ...
                                sum( hcap_ret_t * profiel.hcap(:)',2);
                            ret_t        = economy.stock_returns(:,t) * profiel.stocks(:)' + ret_nonstock;
                        end
                    end
                    ret_t(isnan(ret_t)) = 1;
                    ret_t_monet              = (ret_t-1) .* assets(:,:,t);
                    if par.toed==1 || par.toed==2 % (leeftijdsafhankelijk) toedelen van rendementen
                        % formule (5) in Muns & Werker (feb2019):
                        ret_hedgeFH         = toed_renteafdekking * (annuityreturn_t - 1) + 1;
                        
                        % formule (8) in 'monetaire vorm'
                        ret_hedgeF_monet           = (ret_hedgeFH-1) .* assets(:,:,t);
                        ret_hedgeF_monet(:,Iworkt) = ret_hedgeF_monet(:,Iworkt) + (ret_hedgeFH(:,Iworkt) - toed_hcapafdekking * hcap_ret_t(:,Iworkt)) .* hcap(:,Iworkt);
                        
                        extra_return_monet       = sum( ret_t_monet - ret_hedgeF_monet, 2);
                        if par.toed==1
                            ret_extra_monet_by_age   = (total_wealth_t ./ sum(total_wealth_t,2)) .* extra_return_monet;
                        else
                            if par.toed==2 % geef iedereen zelfde aantal annuteiten obv speculatieve rendement
                                v          = squeeze( annuityprice(:,:,t) - 1 ) * .5;
                                ret_extra_monet_by_age   = (v./sum(v,2)) .* extra_return_monet;
                                clear v;
                            end
                        end
                        ret_t_monet              = ret_extra_monet_by_age + ret_hedgeF_monet;
                    else
                        if par.toed==3
                            v           = squeeze( annuityprice(:,:,t)) * .5;
                            ret_t_monet = (v./sum(v,2)) .* sum(ret_t_monet,2);
                            clear v;
                        end
                    end
                    if par.contr_IGR~=0 && strcmp(par.stock_IGR,'als indiv assets')
                        % return op sol.vermogen is zelfde als return op gehele portefeuille
                        % schaal returns als aandelenexposure > 1:
                        if size(profiel.stocks,2)==1
                            profiel.stocks = profiel.stocks';
                        end
                        stock_exposure = sum(profiel.stocks .* assets(:,:,t), 2) ./ sum( assets(:,:,t), 2);
                        % als er een shortrestrictie op de IGR-buffer is:
                        scale_factor   = 1;
                        gewogen_ret_t  = sum(  (assets(:,:,t) + ret_t_monet./scale_factor), 2) ./ sum( assets(:,:,t), 2);
                        clear stock_exposure scale_factor;
                    end
                    if strcmp(par.contract,'4A') || strcmp(par.contract,'4B')
                        if ~isnan(par.minret)% minimale return, om negatieve financiele assets uit te sluiten
                            % Om -Inf/Inf-uitkomsten te voorkomen: voorkom neg totale assets
                            sumassets0       = max(0, sum(assets(:,:,t) + ret_t_monet,2)); % nodig voor omvang totale assets na return t
                            sumassetsminret  = (1+par.minret) .* sum(assets(:,:,t),2);
                            % Als sumassets0 < sumassetsminret, dan is minimale return sumassetsminret niet waar te maken voor iedereen.
                            % Schaal in dat geval de minimale return naar beneden, zodat deze alsnog gehaald kan worden.
                            vminret          = min(sumassets0 ./ sumassetsminret, 1) * (1+par.minret);
                            sumassetsminret  = vminret .* sum(assets(:,:,t),2);
                            sumassets1       = sum(max(vminret .* assets(:,:,t), assets(:,:,t) + ret_t_monet),2);   % kap af op min.return
                            A_tmp            = vminret.*assets(:,:,t) + max( assets(:,:,t) + ret_t_monet-vminret .* assets(:,:,t),0) ...
                                .* (sumassets0 - sumassetsminret)./(sumassets1 - sumassetsminret);   % schaal totale assets naar beneden bij
                            A_tmp            = squeeze(A_tmp);
                            A_tmp(isnan(A_tmp)) = 0;
                            assets(:,:,t)    = A_tmp;
                            clear sumassets0 sumassets1 sumassetsminret A_tmp;
                        else
                            % ken beleggingsrendementen toe aan persoonlijk pensioenvermogen
                            assets(:,:,t)    = assets(:,:,t) + ret_t_monet;
                        end
                    end
                    if par.contr_IGR~=0
                        % ken rendementen toe aan solidariteitsvermogen
                        if strcmp(par.stock_IGR,'als indiv assets')
                            ret_IGR(:,t) = gewogen_ret_t;
                        else
                            if strcmp(par.portfolio_rebalance,'continu')
                                ret_IGR(:,t) = economy.stock_returns(:,t) .^ par.stock_IGR .* ...
                                    squeeze( economy.bond_returns(:,economy.bond_portf_maturities==par.maturity_IGR,t) ) .^ par.bond_IGR .* ... 
                                    economy.euribor_return(:,t) .^ par.euribor_IGR .* prod( hcap_ret_t(:, Iwork) .^ ( hcap ./ sum(hcap, 2) ) , 2) .^ par.hcap_IGR;
                            else
                                ret_IGR(:,t) = economy.stock_returns(:,t) .* par.stock_IGR + ...
                                    squeeze( economy.bond_returns(:,economy.bond_portf_maturities==par.maturity_IGR,t) ) .* par.bond_IGR + ... 
                                    sum(hcap_ret_t(:,Iwork) .* hcap, 2)./ sum(hcap, 2) .* par.hcap_IGR ;
                            end
                        end
                        assets_IGR(:,t) = ret_IGR(:,t) .* assets_IGR(:,t);
                    end
                end
                if strcmp(par.contract(1),'1') || strcmp(par.contract,'4B')
                    accruals_before_per_contr = squeeze( accrualPVsum(:,t) ./ sum( contr(:,:,t), 2 ));
                    accruals_end_per_contr    = squeeze( accrualPVsum_end(:,t) ./ sum( contr(:,:,t), 2 ));
                    fundratio_0_contr(:,t)    = (fundratio_beforeindex(:,t)-1) .* accruals_before_per_contr;
                    fundratio_1_contr(:,t)    = (fundratio_afterindex(:,t)-1) .* accruals_end_per_contr;
                end
            end
            benefityr(:,:,t1) = sum(benefit_periods(:,par.stepyr:par.stepyr:end,t+1-par.stepyr:t),3);
        end

        if par.AOW > 0 % druk alles uit in termen van gehele loon, dus incl AOW en excl franchise
            benefit_periods      = par.AOW + benefit_periods;
            benefityr            = par.AOW + benefityr;
            pension_wage_profile = pension_wage_profile + profiel.franchise(Iwork);
            profiel.franchise(:) = 0;
        end
        
        % benefit per cohort, schaal door overlevend aantal personen
        vsurv                    = [profiel.Psurv(find(Iwork,1,'last')) profiel.Psurv(Ipension)'];
        benefit_period_coh       = permute(extractDiag(benefit_periods ./ vsurv(1:end-1), 1:3), [1 3 2]) ;
        avg_benefitcoh           = squeeze( nanmean(benefit_period_coh,2));
        repl_rate_benavg_wageavg = squeeze( avg_benefitcoh * par.stepyr ./ nanmean(profiel.wage(Iwork) - profiel.franchise(Iwork)) );
        % verhouding eerste pensioeninkomen tov gemiddelde pensioengevend inkomen
        repl_rate_ben1_wageavg  = squeeze(benefit_periods(:,1,:)) * par.stepyr ./ (mean(profiel.wage(Iwork) - profiel.franchise(Iwork)));
        % verhouding eerste pensioeninkomen tov laatstverdiende pensioengevende loon
        repl_rate_ben1_wagelast = squeeze(benefit_periods(:,1,:)) * par.stepyr ./ (profiel.wage(num_work_periods) - profiel.franchise(num_work_periods));

        % uitkeringen per cohort en per leeftijd
        benefitcohageperiod = NaN(nsim, num_pension_years*par.stepyr, num_cohorts);
        % netto contante waarde: RTS-verdisconteerde som van uitkeringen per scenario en per cohort
        ncw_benefitcoh  = NaN(par.nsim, num_cohorts);
        
        for c=1:num_cohorts % per par.stepyr cohorten
            index_pension_ageper   = 1+(round(v_first_pension_age_coh(c)*par.stepyr):round(v_last_pension_age_coh(c)*par.stepyr))-par.pension_age*par.stepyr;
            index_pension_periods  = v_first_pension_period_coh(c):v_last_pension_period_coh(c);
            benefitcohageperiod(:, index_pension_ageper, c) = benefit_period_coh(:,index_pension_periods,c);
            % verdiscontterde benefits per cohort op pensioendatum
            ncw_benefitcoh(:,c) = squeeze( sum( benefitcohageperiod(:, index_pension_ageper, c) ...
                    .* discount_mat_util(:, 1:length(index_pension_periods), floor(v_first_pension_year_coh(c))), 2) );
        end
        pensioengroei10         = NaN(nsim, num_cohorts);
        if num_pension_periods >= 11*par.stepyr
            % groei pensioen na 10 jaar
            pensioengroei10         = squeeze( benefitcohageperiod(:,11*par.stepyr,:) ./ benefitcohageperiod(:,par.stepyr,:));
        end
        pensioengroei20         = NaN(nsim, num_cohorts);
        if num_pension_periods >= 21*par.stepyr
            % groei pensioen na 20 jaar
            pensioengroei20         = squeeze( benefitcohageperiod(:,21*par.stepyr,:) ./ benefitcohageperiod(:,par.stepyr,:));
        end
        mCRRA        = NaN(nsim, num_cohorts);
        mCRRA_report = NaN(nsim, num_cohorts);           
        for c = 1:num_cohorts
            switch par.utiltype
                case "NCW_CRRA_ben"
                    if c <= num_pension_periods
                        index_pension_ages_c = num_pension_periods+1-v_num_pension_periods_coh(c):num_pension_periods;
                    else
                        index_pension_ages_c = 1:v_num_pension_periods_coh(c);
                    end
                    m_c = benefitcohageperiod(:, index_pension_ages_c, c)/(mean(pension_wage_profile(pension_wage_profile>0)) * v_num_work_periods_coh(c) / num_work_periods);
                    mCRRA(:,c)        = CRRAutility( m_c, par.gamma,        par.bCEQ, discount_mat_util(:, 1:v_num_pension_periods_coh(c), v_first_pension_period_coh(c)), false, true);
                    mCRRA_report(:,c) = CRRAutility( m_c, par.gamma_report, par.bCEQ, discount_mat_util(:, 1:v_num_pension_periods_coh(c), v_first_pension_period_coh(c)), false, true);
                case "CRRA_NCW_ben"
                    mCRRA(:,c)        = CRRAutility( ncw_benefitcoh(:,c) , par.gamma       , par.bCEQ, [], true, true );
                    mCRRA_report(:,c) = CRRAutility( ncw_benefitcoh(:,c) , par.gamma_report, par.bCEQ, [], true, true );
                case "CRRA_avgben"
                    mCRRA(:,c)        = CRRAutility( squeeze( nanmean(benefit_period_coh(:,:,c),2)) * par.stepyr, par.gamma       , par.bCEQ, [], true, true );
                    mCRRA_report(:,c) = CRRAutility( squeeze( nanmean(benefit_period_coh(:,:,c),2)) * par.stepyr, par.gamma_report, par.bCEQ, [], true, true );
                case "assets/accruals pensioendatum"
                    if strcmp(par.contract(1),'1') || strcmp(par.contract,'4B')
                        % accrualPVsum_pension
                        v = annuityprice(:,num_work_periods+1,v_first_pension_period_coh(c)+1) .* accrual(:,num_work_periods,v_first_pension_period_coh(c));
                    else
                        v = assets(:,num_work_periods,v_first_pension_period_coh(c));
                    end
                    mCRRA(:,c)              = CRRAutility( v, par.gamma       , par.bCEQ, [], true, true );
                    mCRRA_report(:,c)       = CRRAutility( v, par.gamma_report, par.bCEQ, [], true, true );
                    clear v;
            end
        end
        
        fraction_IGRbuffer_contr  = [];
        fraction_IGRbuffer_sumassets  = [];
        assets1                   = [];
        assets_end                = [];
        if par.contr_IGR~=0 && ~isnan(par.contr_IGR)
            % fractie in IGR als % jaarlijkse premiesom inclusief IGR-premie
            fraction_IGRbuffer_contr = assets_IGR ./ squeeze( sum( contr, 2 )) ;
            % fractie in IGR als % assets exclusief IGR-buffer (net zoals dkg-perc zonder buffer is)
            fraction_IGRbuffer_sumassets = assets_IGR ./ squeeze( sum( assets, 2 )) ;
        end
        sum_assets = 0;
        if ~isempty(assets)
            sum_assets = squeeze(sum(assets,2));
        end
        if ~isempty(assets_sol)
            sum_assets = sum_assets + assets_sol;
        end
        if ~isempty(assets_IGR)
            sum_assets = sum_assets + assets_IGR;
        end
        if sum_assets==0
            sum_assets=[];
        end
        if strcmp(par.contract(1),'4')
            assets1                   = squeeze( assets(:,1,:) );
            assets_end                = squeeze( assets(:,:,end) );
        end
        
        [avg_benefitcoh_0, ncw_benefitcoh_0, pensioengroei10_0, pensioengroei20_0, mCRRA_0, mCRRA_report_0,...
            indexation_0, fundratio_beforeindex_0, fundratio_afterindex_0, fundratio_0_contr_0, fundratio_1_contr_0, fraction_IGRbuffer_sumassets_0, fraction_IGRbuffer_contr_0, ...
            mexposure1_0, mexposure2_0, repl_rate_benavg_wageavg_0, repl_rate_ben1_wageavg_0, repl_rate_ben1_wagelast_0, assets1_0, assets_end_0, assets_sol_0, sum_assets_0] = ...
        add_rows({avg_benefitcoh_0, ncw_benefitcoh_0, pensioengroei10_0, pensioengroei20_0, mCRRA_0, mCRRA_report_0, indexation_0, fundratio_beforeindex_0, ...
            fundratio_afterindex_0, fundratio_0_contr_0, fundratio_1_contr_0, fraction_IGRbuffer_sumassets_0, fraction_IGRbuffer_contr_0, mexposure1_0, mexposure2_0, ...
            repl_rate_benavg_wageavg_0, repl_rate_ben1_wageavg_0, repl_rate_ben1_wagelast_0, assets1_0, assets_end_0, assets_sol_0, sum_assets_0}, ...
            {avg_benefitcoh, ncw_benefitcoh, pensioengroei10, pensioengroei20, mCRRA, mCRRA_report, indexation, fundratio_beforeindex, fundratio_afterindex, fundratio_0_contr, fundratio_1_contr,...
            fraction_IGRbuffer_sumassets, fraction_IGRbuffer_contr, squeeze(mexposure(:,:,1)), squeeze(mexposure(:,:,2)), repl_rate_benavg_wageavg, repl_rate_ben1_wageavg, ...
            repl_rate_ben1_wagelast, assets1, assets_end, assets_sol, sum_assets});
        end
        clear ncw_benefitcoh  pensioengroei10  pensioengroei20 indexation fundratio_beforeindex fundratio_afterindex ...
             mexposure repl_rate_benavg_wageavg  repl_rate_ben1_wageavg  repl_rate_ben1_wagelast assets
        % laatste cohort waarvan gehele pensioenperiode bekend is:
        c0      = num_cohorts - num_pension_periods + 1;
        saved   = false;
        if bmaxUtil_ben || bmaxUtil_ncwben || bmaxUtil_avgben % maximaliseer nut mbt stock exposure SV,IGR
            if bmaxUtil_ncwben % U = u(NCW benefits) (niet in CEQ)
                vutil(iter) = mean(CRRAutility(ncw_benefitcoh_0(:,c0), par.gamma, par.bCEQ, [], true, true));
            end
            if bmaxUtil_ben
                vutil(iter) = mean( mCRRA_0(:,c0) );
            end
            if bmaxUtil_avgben
                vutil(iter) = mean(CRRAutility( avg_benefitcoh_0(:,c0) , par.gamma, par.bCEQ, [], true, true));
            end
            
            fprintf([num2str(iter) '-'])
            if isnan(vstocks_iter(iter+1))
                [maxUtil,Imax] = max(vutil(1:iter));
                opt_stock  = vstocks_iter(Imax);
                if iter < 30
                    try
                        vstocks_tmp(1) = max(vstocks_iter( vstocks_iter(1:iter) < opt_stock ) );
                    catch
                        vstocks_tmp(1) = 0;
                    end
                    try
                        vstocks_tmp(2) = min(vstocks_iter( vstocks_iter(1:iter) > opt_stock ) );
                    catch
                        vstocks_tmp(2) = 1;
                    end
                    vstocks_new     = setdiff( linspace( vstocks_tmp(1) , ...
                            vstocks_tmp(2),7), vstocks_iter(1:iter));
                    % drop already known elements:
                    vstocks_iter(iter+(1:length(vstocks_new))) = vstocks_new;
                    disp(' ')
                    disp([num2str(maxUtil) ':' num2str(vstocks_new)])
                    disp(' ')
                    iter = iter + 1;
                else % stop na 30 bisectie-iteraties
                    disp('')
                    disp('optimale exposure aandelen solidariteitsvermogen')
                    disp(opt_stock)
                    bcontinue     = false;
                    saved = simulatie(Excelfile,sdir,sheetcell{i},opt_stock,stat_col);
                end
            else
                iter = iter + 1;
            end
        else
            % let op: eerste twee periodes wordt niets uitgekeerd omdat eerste periode een
            % 'dummyjaar' is zodat t-1 gebruikt kan worden bij initile
            % periode t=2. Geboortecohort is nu herrekend, alsof in jaar 1 wordt gestart.
            c0text  = ['geboortecohort ' num2str(c0 - 1 - par.max_age)];
            bcontinue = false; % geen optimalisatie -> stop
        end
    end
    if ~saved
        if isempty(mexposure1_0) && isfield(profiel,'stocks')
            mexposure1_0 = profiel.stocks;
            mexposure2_0 = profiel.bond;
        end
        coh_weight = (min([v_num_work_periods_coh/num_work_periods ; v_num_pension_periods_coh/num_pension_periods]) .* par.disconto_coh.^(1:num_cohorts))'; % weeg cohort obv discontovoet
        writeDescriptives(stat_col,inputfile,par,sheetcell{i},c0,ncw_benefitcoh_0, pensioengroei10_0, pensioengroei20_0, ...
            mCRRA_0, mCRRA_report_0, v_first_pension_period_coh, repl_rate_benavg_wageavg_0, repl_rate_ben1_wageavg_0, repl_rate_ben1_wagelast_0, ...
            indexation_0, fundratio_beforeindex_0, fundratio_afterindex_0, ...
            fundratio_0_contr_0, fundratio_1_contr_0, fraction_IGRbuffer_sumassets_0, ...
            fraction_IGRbuffer_contr_0,assets_sol_0,sum_assets_0,assets1_0,assets_end_0,mexposure1_0, mexposure2_0, coh_weight);
        saved = true;
        if ~strfind(version,'2017a')
        % Matlab geeft foutmelding in 2017a als openGL-optie niet aanstaat
            makeFigures(v_birth_year_coh, pensioengroei10_0, pensioengroei20_0, repl_rate_benavg_wageavg_0, ...
                repl_rate_ben1_wageavg_0, repl_rate_ben1_wagelast_0, ncw_benefitcoh_0, num_pension_periods, par, c0, c0text)
        end
    end
    vutil = [];
    end
end
end

function varargout = add_rows(A,B)
% voeg rijen in B toe aan A
    varargout = cell(1,length(A));
    for i = 1:length(A)
        varargout{i} = [A{i}; B{i}];
    end

end

function writeDescriptives(stat_col,dirfile, par, sheet,c0,ncw_benefitcoh, pensioengroei10, pensioengroei20, ...
        mCRRA, mCRRA_report, ... 
        v_first_pension_period_coh, repl_rate_benavg_wageavg, repl_rate_ben1_wageavg, repl_rate_ben1_wagelast, ...
        indexation, fundratio_0, fundratio_1, fundratio_0_contr, ...
        fundratio_1_contr, fraction_IGRbuffer_sumassets, fraction_IGRbuffer_contr, ...
        assets_sol,sum_assets,assets1,assets_end, stockexposure, bondexposure, coh_weight)
    
    colLetters = xlsColNum2Str(1:500); % converteer kolomnummer naar kolomletter  
    mM = [ncw_benefitcoh(:,c0)          pensioengroei10(:,c0) pensioengroei20(:,c0) ...
            mCRRA(:,c0)  mCRRA_report(:,c0) ...
            repl_rate_benavg_wageavg(:,v_first_pension_period_coh(c0)) ...
            repl_rate_ben1_wageavg(:,v_first_pension_period_coh(c0)) ...
            repl_rate_ben1_wagelast(:,v_first_pension_period_coh(c0))];
    cvars = {'ncw_benefit','pensgroei10','pensgroei20', ... 
        ['CRRA_' num2str(par.gamma) '_benefits'], ['CRRA_' num2str(par.gamma_report) '_benefits'], ... 
        'vvr_gempensgemloon', 'vvr_gemloon','vvr_laatsteloon'};
    [Mstat,rowns] = displayDescriptives(mM,cvars, [],par.nsim,[],[false true true false false true true true], ...
        [false true true false false false false false]);
    startcol  = 15;
    vheading  = ['geboortecohort' cvars NaN cvars];
    cMstat    = [NaN vheading(1+(1:length(cvars))); rowns(:) num2cell(Mstat) ];
    [col2,row2a] = xlswrite_wrap(dirfile, cMstat, sheet, startcol, 1, colLetters, true);
    
    % CRRA-resultaten wegschrijven naar tabblad stats:
    xlswrite_wrap(dirfile, {sheet}, 'stats',stat_col,2,colLetters, true);
    I     = contains(cvars,'CRRA_');
    Mstat = Mstat(:, I);
    vgam0 = cvars(I);
    vgam  = cell(size(vgam0));
    vrowns = [];
    for i=1:length(vgam0)
        stmp    = vgam0{i};
        vgam{i} = ['gamma = ' stmp(6)];
        rowns(end-1:end) = {vgam{i}; NaN}; % geen pos en neg rij nodig, wel lege rij
        vrowns  = [vrowns; rowns(:)];
    end
    Mstat(end-1:end,:) = NaN;  % geen pos en neg rij nodig, wel lege rijen
    I          = ~any(isnan(mCRRA));
    weightCRRA = mean(mCRRA(:,I) * coh_weight(I)/sum(coh_weight(I)));
    
    if size(stockexposure,2)==1
        stockexposure = stockexposure';
    end
    if size(bondexposure,2)==1
        bondexposure  = bondexposure';
    end
    if (par.contract(1)=='1') || strcmp(par.contract,'4B')
        w_assets = [par.stock_sol; par.bond_sol; par.euribor_sol];
        if strcmp(par.contract,'4B')
            w_sol = assets_sol(:,end) ./ (assets_sol(:,end) + sum(assets_end,2));
            w_assets(1) = mean( w_assets(1) * w_sol + (1 - w_sol) .* nansum( stockexposure .* assets_end , 2) ./ nansum( assets_end, 2) );
            w_assets(2) = mean( w_assets(2) * w_sol + (1 - w_sol) .* nansum( bondexposure  .* assets_end  , 2) ./ nansum( assets_end, 2) );
            w_assets(3) = mean( w_assets(3) * w_sol + (1 - w_sol) .* nansum( (1 - stockexposure - bondexposure) .* assets_end , 2) ./ nansum( assets_end, 2) );
        end
    else
        % gewichten in termen van assets
        w_assets      = NaN(size(assets_end,1),3);
        % stocks:
        w_assets(:,1) = sum( stockexposure .* assets_end , 2) ./ sum( assets_end, 2);
        % bonds:
        w_assets(:,2) = sum( bondexposure .* assets_end , 2) ./ sum( assets_end, 2);
        % euribor:
        w_assets(:,3) = 1 - sum(w_assets(:,1:2),2);
        w_assets      = mean(w_assets)';
    end
    if size(stockexposure,1)==1 % terug transformeren
        stockexposure = stockexposure';
        bondexposure  = bondexposure';
    end
        
    if par.contract(1)=='1'
        bond_maturity = par.maturity_sol;
    else
        bond_maturity = par.optLC_bondmat;
    end
    if ~isnumeric(bond_maturity)
        bond_maturity = NaN;
    end
    if strcmp(par.stock_IGR,'als indiv assets')
        w_stock_IGR = w_assets(1);
    else
        w_stock_IGR = par.stock_IGR;
    end
    xlswrite_wrap(dirfile, [Mstat(:); w_assets; par.stock_sol; w_stock_IGR; bond_maturity; weightCRRA], 'stats', stat_col, 3, colLetters, true);
    % voeg definitie van rijen toe
    xlswrite_wrap(dirfile, {['geboortecohort ' num2str(c0-2-par.max_age)]}, 'stats',1,1,colLetters, true);
    xlswrite_wrap(dirfile, [vrowns; 'stocks_assets '; 'bonds_assets  '; 'euribor_assets'; 'sol_stocks    '; 'IGR_stocks    '; 'bond_maturity '; 'wghtCRRA      '], 'stats', 1, 3, colLetters, true);
    
    cvars_per = {};
    if ~isempty(fundratio_1_contr)
        mMbuffer_per = fundratio_1_contr(:,end);
        cvars_per = [cvars_per {'excess funding ratio % contribution'}];
    else
        mMbuffer_per = zeros(size(fraction_IGRbuffer_contr, 1),0);
    end
    if ~isempty(fraction_IGRbuffer_contr)
        mMbuffer_per = [mMbuffer_per  fraction_IGRbuffer_contr(:,end)];
        cvars_per = [cvars_per {'IGR funding % contribution'}];
    end
    if ~isempty(fraction_IGRbuffer_sumassets)
        mMbuffer_per = [mMbuffer_per  fraction_IGRbuffer_sumassets(:,end)];
        cvars_per = [cvars_per {'IGR funding % sum assets'}];
    end
    if ~isempty(mMbuffer_per)
        [Mstat_per,rowns] = displayDescriptives(mMbuffer_per,cvars_per, [],par.nsim,[],true(1,size(mMbuffer_per,2)), ...
            false(1,size(mMbuffer_per,2)));
        vheadingper = [NaN cvars_per];
        cMstat      = [NaN vheadingper(1+(1:length(cvars_per))); rowns(:) num2cell(Mstat_per) ];
        [~,row2b]   = xlswrite_wrap(dirfile, cMstat, sheet, col2+2, 1, colLetters, true);
        row2        = max([row2a row2b]) + 2;
    else
        row2 = row2a + 2;
    end
    
    mM = NaN(length(v_first_pension_period_coh), length(vheading));
    for c = 1:size(mCRRA,2)
        mMc = [ncw_benefitcoh(:,c)          pensioengroei10(:,c) pensioengroei20(:,c) ...
                mCRRA(:,c)  mCRRA_report(:,c) ...
                repl_rate_benavg_wageavg(:,c) ...
                repl_rate_ben1_wageavg(:,v_first_pension_period_coh(c)) ...
                repl_rate_ben1_wagelast(:, v_first_pension_period_coh(c))];
        mM(c,:) = [(c-1)/par.stepyr-par.max_age-1  mean(mMc) NaN median(mMc)];
    end
    cMstat   =[cell(1) 'mean'          cell(size(cvars)) 'median' cell(size(cvars)-[0 1]) ;
               vheading; num2cell(mM( par.stepyr*(par.max_age+1-par.pension_age):end+1-(par.stepyr*(par.max_age+1-par.pension_age)),:)) ];
    currcol = xlswrite_wrap(dirfile, cMstat, sheet, startcol, row2, colLetters, true);
    currcol = xlswrite_series(mCRRA(:,par.stepyr*(par.max_age+1-par.pension_age):end),true, par.nsim , ['CRRA' num2str(par.gamma)],...
            'geboortecohort',0-par.pension_age-1,dirfile,sheet,currcol+2, row2, colLetters,par.stepyr);
    if ~isempty(indexation)
        currcol  = xlswrite_series(indexation(:,2:end), true,par.nsim,'indexatie',...
            'periode',0,dirfile,sheet,currcol+2, row2, colLetters, par.stepyr);
        currcol  = xlswrite_series(cumprod(indexation(:,2:end),2),true, par.nsim,'cumulatieve indexatie',...
            'periode',0,dirfile,sheet,currcol+2, row2, colLetters, par.stepyr);
    end
    if ~isempty(fundratio_0)
        currcol  = xlswrite_series(fundratio_0(:,2:end),true, par.nsim,'dekkingsgraad voor indexatie en korting',...
            'periode',0,dirfile,sheet,currcol+2, row2, colLetters, par.stepyr);
    end
    if ~isempty(fundratio_1)
        currcol  = xlswrite_series(fundratio_1(:,2:end),true, par.nsim,'dekkingsgraad na indexatie en korting',...
            'periode',0,dirfile,sheet,currcol+2, row2, colLetters, par.stepyr);
    end
    if any(fundratio_0_contr(:))
        currcol  = xlswrite_series(fundratio_0_contr(:,2:end),true, par.nsim,'excess funding before indexation % premiesom',...
            'periode',0,dirfile,sheet, currcol+2, row2, colLetters, par.stepyr);
    end
    if any(fundratio_1_contr(:))
        currcol  = xlswrite_series(fundratio_1_contr(:,2:end),true, par.nsim,'excess funding after indexation % premiesom',...
            'periode',0,dirfile,sheet, currcol+2, row2, colLetters, par.stepyr);
    end
    if any(fraction_IGRbuffer_contr(:))
        currcol  = xlswrite_series(fraction_IGRbuffer_contr(:,2:end),true, par.nsim,'fractie IGR-buffer % premiesom',...
            'periode',0,dirfile,sheet, currcol+2, row2, colLetters, par.stepyr);
    end
    if any(fraction_IGRbuffer_sumassets(:))
        currcol  = xlswrite_series(fraction_IGRbuffer_sumassets(:,2:end),true, par.nsim,'fractie IGR-buffer % som ind assets',...
            'periode',0,dirfile,sheet, currcol+2, row2, colLetters, par.stepyr);
    end
    if any(sum_assets(:))
        currcol  = xlswrite_series(sum_assets(:,2:end),false, par.nsim,'som assets',...
            'periode',0,dirfile,sheet, currcol+2, row2, colLetters, par.stepyr);
    end
    if any(assets1(:))
        rowns = {'geomean','mean','stdev','2.5% mean low','97.5% mean high','min','0.5%','1%','2.5%','25%', ...
                 '50%: median','75%','97.5%','99%','99.5%','max','P(<0)','P(>0)'};
        currcol  = xlswrite_series(assets1(:,2:end),false, par.nsim,'assets na eerste werkzame periode',...
            'periode',0,dirfile,sheet, currcol+2, row2, colLetters, par.stepyr,rowns);
    end
    if any(assets_end(:))
        rowns = {'geomean','mean','stdev','2.5% mean low','97.5% mean high','min','0.5%','1%','2.5%','25%', ...
                 '50%: median','75%','97.5%','99%','99.5%','max','P(<0)','P(>0)'};
        currcol  = xlswrite_series(assets_end,false, par.nsim,'assets in laatste simulatieperiode',...
            'na leeftijd',par.first_exposure_age,dirfile,sheet, currcol+2, row2, colLetters,par.stepyr,rowns);
    end
    if any(stockexposure(:))
        currcol  = xlswrite_series(repmat(stockexposure',par.nsim,1),true, par.nsim,'stock exposure in laatste simulatieperiode',...
            'leeftijd',par.first_exposure_age,dirfile,sheet, currcol+2, row2, colLetters,par.stepyr);
    end
end

function B = B_BandX(vtau,a,Delta)
% function B, eq (4), Brennan and Xia (2000)
% vtau = looptijden, a = par.V_alp, Delta = 1/par.stepyr
%   B      = (1 - exp(-par.V_alp * economy.RTS_maturities))/par.V_alp;
%     B(t) = 1/D*\int_(s=t-D)^t B(s) ds
%          = 1/(Da)*\int_(s=t-D)^t [1-exp(-as) ] ds
%          = 1/(Da)*[s+1/a*exp(-as) ]_(s=t-D)^t
%          = 1/a*[1+exp(-at)/(Da)*(1-exp(aD) )]
%    met a = par.V_alp, D = 1/par.stepyr
    B      = 1/a * (1 + exp(-a*vtau) / (a*Delta) .* (1-exp(a*Delta)) );
end

function currcol  = xlswrite_series(data,bperc,numobs,sname,stype,startrow,dirfile,sheet,currcol, rownr, colLetters,stepyr,I)
    nsim_hor = size(data,2);
    [Mstat,rownms] = displayDescriptives(data, [],[], numobs, [], ...
        repmat(bperc,1,nsim_hor),false(1,nsim_hor),false);
    if ~exist('I','var') || isempty(I)
        I = 1:length(rownms)-2;
    else
        I = ismember(rownms,I);
    end
    cMstat   =[sname      cell(1,length(rownms(I))) ;
           stype      rownms(I);
           num2cell([startrow+1/stepyr*(0:size(Mstat,2)-1)'  Mstat(I,:)']) ];
    currcol  = xlswrite_wrap(dirfile, cMstat, sheet, currcol, rownr, colLetters, true);        
end

function U = CRRAutility(v, gamma, bCEQ, weight, bCRRAscen, skip0)
%CRRAutility bepaald CRRA utility voor de uitkomsten in vector v 
% v: elementen gebruikt om CEQ te bepalen bij nut in gamma
% gamma: risicoaversie
% bCEQ = 0: utility, 1: CEQ per scenario, 2: CEQ over alle scenario's
% weight bevat evt periodegewichten
% bCRRAscen = true als CRRA over scenario's als n obs per scenario
% skip0 = true als 0 niet wordt genomen, zodat NaN's worden vermeden
    if ~exist('bCRRAscen','var')
        bCRRAscen = true;
    end
    if size(v,2)==1 && isnumeric(gamma) && bCRRAscen
        v = v'; % CRRA toegepast over scenario's
    end
    if ~exist('weight','var') || isempty(weight)
        weight = ones(size(v));
    else
        if sum(size(weight)==1)>0 && bCEQ~=2
            weight = repmat(weight(:)', [size(v,1) 1] );
        end
    end
    if exist('skip0','var') && skip0
        weight      = weight .* ((v>0) & ~isnan(v));
        weight_corr = mean((v>0) | isnan(v)); % neglect nonpositive assets
    else
        weight_corr = 1;
    end
    if length(gamma)>1 || ~exist('bCEQ','var') || isempty(bCEQ)
        bCEQ = 1;
    end
    if (gamma~=1)
        U = nansum( weight .* weight_corr .* (repmat(v,[1 size(weight,2)/size(v,2)]).^(1-gamma) - 1), 2) / (1-gamma);
    else
        U = nansum( weight .* weight_corr .*repmat(log(v),[1 size(weight,2)/size(v,2)]), 2);
    end
    if bCEQ % reken nut in y om naar certainty equivalents
        if bCEQ==2 % CEQ obv gemiddelde over alle scenario's
            if (gamma~=1)  
                % mean(U) = mean(sum( weight .* (CEQ.^(1-gamma) - 1), 2) / (1-gamma))
                % mean(U)*(1-gamma) = mean(sum( weight, 2)) .* (CEQ.^(1-gamma) - 1)
                % mean(U)*(1-gamma) ./ mean(sum( weight, 2)) = (CEQ.^(1-gamma) - 1)
                % (mean(U)*(1-gamma) ./ mean(sum( weight, 2)) + 1).^(1/(1-gamma)) = CEQ
                U = repmat( ( mean(U)*(1-gamma) ./ mean(sum( weight, 2)) + 1).^(1/(1-gamma)), [size(v,1) 1]);
            else
                % mean(U) = mean(sum( weight .* ln(CEQ), 2)
                % mean(U) = mean(sum( weight, 2)) * ln(CEQ)
                % exp(mean(U) ./ mean(sum( weight, 2)) ) = CEQ
                U = repmat( exp(mean(U) ./ mean( sum( weight, 2)) ), [size(v,1) 1]);
            end            
        else  % CEQ per scenario
            if (gamma~=1)
                % U = sum( weight .* (CEQ.^(1-gamma) - 1), 2) / (1-gamma)
                % U*(1-gamma) = sum( weight, 2) .* (CEQ.^(1-gamma) - 1)
                % U*(1-gamma) ./ sum( weight, 2) = (CEQ.^(1-gamma) - 1)
                % (U*(1-gamma) ./ sum( weight, 2) + 1).^(1/(1-gamma)) = CEQ
                U = (U*(1-gamma) ./ sum( weight, 2) + 1).^(1/(1-gamma));
            else
                % U = sum( weight .* ln(CEQ), 2)
                % U = sum( weight, 2) * ln(CEQ)
                % exp(U ./ sum( weight, 2) ) = CEQ
                U = exp(U ./ sum( weight, 2) );
            end
        end
    end
end

function makeFigures(v_birth_coh, pensioengroei10, pensioengroei20, repl_rate_benavg_wageavg, ...
    repl_rate_ben1_wageavg, repl_rate_ben1_wagelast, ncw_benefitcoh, num_pension_years, par,c0, c0text)
    %makeFigures maak figuur van groei pensioenuitkering, vervangingsratio
    % en som verdisconteerde uitkeringen.
    figure;
    subplot(1,2,1)
    [AX,~,~] = plotyy(v_birth_coh(num_pension_years:end), ...
        100*[median(pensioengroei10(:,num_pension_years:end)); median(pensioengroei20(:,num_pension_years:end)); ...
        median(repl_rate_benavg_wageavg(:,num_pension_years:end)); ...
        median(repl_rate_ben1_wageavg); ...
        median(repl_rate_ben1_wagelast)]', ...
            v_birth_coh(num_pension_years:end), median(ncw_benefitcoh(:,num_pension_years:end)));
    
        set(AX(1),'XLim', v_birth_coh([num_pension_years end]));
        set(AX(2),'XLim', v_birth_coh([num_pension_years end]));
        xlabel('geboorteperiode')
        ylabel('%')
        legend(['uitkering ' num2str(par.pension_age + 10) '/' num2str(par.pension_age)] , ...
                ['uitkering ' num2str(par.pension_age + 20) '/' num2str(par.pension_age)], ...
                'vvr: gem.uitkering/laatste loon', ...
            ['vvr: uitkering ' num2str(par.pension_age) '/gem.loon'], ...
            ['vvr: uitkering ' num2str(par.pension_age) '/laatste loon'], ...
                'NCW uitkeringen (rechter as)','Location','SouthOutside')
        title('mediane uitkomstmaten')
        
    subplot(1,2,2)
        xl     = prctile(repl_rate_ben1_wageavg(:, c0), [10 90]);
        xl(1)  = min(xl(1),0);
        xl(2)  = max(xl(2),5);
        edges  = linspace(xl(1), xl(2), 100);
        [n, ~] = histc(repl_rate_ben1_wageavg(:, c0), edges );
        bar(edges,n,'histc')
        title(['uitkering ' num2str(par.pension_age) '/gem.loon, ' c0text])
        xlim(xl)
end


function annuityprice0 = getAnnuityPrices(Psurv, discount_grossrate_cum, num_working_periods, growthrate)
    % Psurv                  = overlevingskansen per afzonderlijke periode
    % discount_grossrate_cum = disconteringsmatrix
    % num_working_periods    = aantal werkzame jaren
    % growthrate             = groeivoet van uitbetalingen tijdens pensioenperiode, 0 = constante uitbetalingen
    if ~exist('growthrate','var')
        growthrate = 1;
    end
    num_periods        = length(Psurv);
    [nsim, ~, num_sim_years] = size(discount_grossrate_cum);
    % aantal pensioenjaren
    num_pension_periods     = num_periods - num_working_periods;
    annuityprice0           = NaN(nsim, num_periods, num_sim_years);
    Psurv0                  = [1; Psurv(:)]; % start-of-period overlevingskansen
    if length(growthrate)==1 % geen periode- en leeftijdgevoeligheid
        vbenefit                = [zeros(1,num_working_periods) ones(1,num_pension_periods)];
        temp_mat0               = Psurv0(1:end-1)' .* vbenefit(:)'; % benefits worden aan het begin van de periode betaald.
    end
    % prijs pensioenannuteit per leeftijdsperiode
    for age=1:num_periods
        if length(growthrate)>1
            clear vbenefit num_pension_periods_to_pension_age;
            num_periods_to_pension_age = num_working_periods+1-age;
            growthrate0 = (growthrate(:,max(num_periods_to_pension_age,0)+1,:));
            vbenefit(:,num_working_periods+max(-num_periods_to_pension_age,0)+(1:num_pension_periods+min(0,num_periods_to_pension_age)),1:num_sim_years) ...
                = (growthrate(:,max(num_periods_to_pension_age,0)+(1:min(num_pension_periods,num_periods+1-age)),:) ./ growthrate0);
            temp_mat0               = Psurv0(1:end-1)' .* vbenefit; % benefits worden aan het begin van de periode betaald.
        end
        % in zelfde periode volgt nog evt indexatie over huidige periode
        annuityprice0(:,age,:) = sum( temp_mat0(:,age:num_periods,:)  ...
            .* discount_grossrate_cum(:, 1:num_periods+1-age, :), 2) / Psurv0(age);
    end
end

function [Mstat,rowns] = displayDescriptives(mM,colns,rowns,numsim,vperc0,bperc,bgross,bdisplay)
    % mM     contains the data
    % colns  contains column names
    % rowns  contains row names
    % numsim contains number of observations/simulations
    % vperc  vector that contains percentiles
    % bperc  boolean vector that indicates if a variable is in percentages or not
    % bgross equals one if 1 + rate
    % bdisplay is true if statistics are displayed on screen (default is false)
    if ~exist('colns','var') || isempty(colns)
        colns = cellstr([repmat('t',[size(mM,2) 1]) num2str((1:size(mM,2))')])';
        colns = strrep(colns,' ','');
    end
    if ~exist('rowns','var') || isempty(rowns)
        rowns = {'geomean','mean','stdev','2.5% mean low','97.5% mean high','min','0.5%','1%','2.5%','25%', ...
                 '50%: median','75%','97.5%','99%','99.5%','max','P(<0)','P(>0)'};
    end
    if ~exist('vperc0','var') || isempty(vperc0)
        vperc0 = [0.5 1 2.5 25 50 75 97.5 99 99.5];
    end
    if ~exist('bperc','var') || isempty(bperc)
        bperc = true(1,size(mM,2));
    end
    if ~exist('bgross','var') || isempty(bgross)
        bgross = false(1,size(mM,2));
    else
        if length(bgross)==1
            bgross = repmat(bgross,1,size(mM,2));
        end
    end
    if ~exist('bdisplay','var') || isempty(bdisplay)
        bdisplay = false;
    end
    gm = NaN(1,size(mM,2));
    for i=1:size(mM,2)
        v     = mM(:,i);
        try
            gm(i) = geomean(v(~isnan(v)))-bgross(i);
        catch
            gm(i) = NaN; % geen meetkundig gemiddelde als bijvoorbeeld elementen negatief zijn.
        end
    end
    mM = mM - repmat(bgross, size(mM,1), 1);
    if size(mM,2)==1
        vperctile = prctile(mM, vperc0(:)')';
    else
        vperctile = prctile(mM, vperc0(:)');
    end
    try % nieuwere Matlab-versies met omitnan-optie
        m0 = mean(mM,'omitnan');
        s0 = std(mM,'omitnan');
        Mstat = [gm; m0; s0; m0+norminv(.025)*s0/numsim^.5; m0+norminv(.975)*s0/numsim^.5; min(mM); ...
            vperctile; max(mM); mean(mM<0); mean(mM>0)];
    catch % oudere Matlab-versies met NaN-functie
        m0 = nanmean(mM);
        s0 = nanstd(mM);
        Mstat = [gm; nanmean(mM); nanstd(mM); m0+norminv(.025)*s0/numsim^.5; m0+norminv(.975)*s0/numsim^.5; ...
            min(mM); vperctile; max(mM); mean(mM<0); mean(mM>0)];
    end
    perctable = repmat(100.^(bperc(:)'),[size(Mstat,1) 1]);
    Mstat     = perctable .* Mstat;
    if bdisplay
        disp(array2table(Mstat,'VariableNames',colns,'RowNames',rowns));
    end
end

function economy0 = readKNWset(spath,par,vmatur)
% lees scenario's in die met Matlab zijn gegenereerd
    if ~isempty(par.sfile)
        load([spath par.sfile]);
    end
    nsim_hor2   = par.stepyr*par.nsim_hor+2;
    maxmaturity = 100;
    all_ones1 = ones(par.nsim, nsim_hor2);
    if isnumeric(par.optLC_bondmat) && ~isnan(par.optLC_bondmat)
        vmatur = unique([vmatur par.optLC_bondmat]);
    end
    if isnumeric(par.maturity_sol) && ~isnan(par.maturity_sol)
        vmatur = unique([vmatur par.maturity_sol]);
    end
    if strcmp(par.scenarioset,'all 1') % scenario with all ones
        % vlakke RTS die over de tijd varieert. Hier: rente gelijk aan 10-jaarsrente
        economy0.stock_returns    = all_ones1;
        economy0.bond_portf_maturities   = vmatur;
        economy0.bond_returns     = ones(par.nsim, length(economy0.bond_portf_maturities), nsim_hor2);
        economy0.euribor_return   = all_ones1;
        economy0.price_inflation  = all_ones1;
        economy0.wage_inflation   = all_ones1;
        economy0.RTS              = ones(par.nsim, 1+par.stepyr*maxmaturity, nsim_hor2);
        economy0.RTS_maturities   = linspace(0,maxmaturity,1+par.stepyr*maxmaturity);
        return
    end
    if strcmp(par.set_RTS,'scenarioset')
        [economy0.RTS,economy0.RTS_maturities] = constructRTS(economy);
        economy0.RTS             = 1+economy0.RTS(1:par.nsim,:,1:nsim_hor2);
        
        economy0.euribor_return  = NaN(par.nsim, nsim_hor2);
        economy0.bond_portf_maturities   = economy.bond_portf_maturities;
        economy0.bond_returns    = NaN(par.nsim, length(economy0.bond_portf_maturities), nsim_hor2);
        if strcmp(par.bond_rebalance,'continu') % herbalanceren maakt alleen uit bij variabele rente
            economy0.euribor_return(:,2:end) = 1+economy.return_euribor(1:par.nsim, 1:nsim_hor2-1);
            economy0.bond_returns(:,:,2:end) = 1+economy.bond_returns(1:par.nsim, :, 1:nsim_hor2-1);
        else
            economy0.euribor_return(:,2:end) = 1+economy.return_euribor(1:par.nsim, 1:par.nsim_hor);
            RTSlocs                          = find(ismember(economy0.RTS_maturities,economy0.bond_portf_maturities(2:end)));
            % looptijd nul is altijd continu tegen euribortarief:
            economy0.bond_returns(:,1,2:end)     = 1+economy0.euribor_return(1:par.nsim, 1, 1:nsim_hor2-1);
            % langere looptijden worden iedere periode geherbalanceerd
            economy0.RTS(1:par.nsim, 1, 1:nsim_hor2-1) = 1; % voor de berekeningen van obligatie met looptijd van n periode
            economy0.bond_returns(:,2:end,2:end) = economy0.RTS(:,RTSlocs,1:end-1) ...
                                                    .^repmat(economy0.bond_portf_maturities(2:end),[par.nsim 1 nsim_hor1-1]) ...
                                                   ./ economy0.RTS(:,RTSlocs-1,2:end) ...
                                                    .^repmat(economy0.bond_portf_maturities(2:end)-1/par.stepyr,[par.nsim 1 nsim_hor1-1]);
        end
    elseif strcmp(par.set_RTS,'Vasicek')
        % construeer RTS-scenarios met Vasicek (JFE1977)-model (scenarios zijn onafh van Black-Scholes)
        % simuleer short rate
        r0          = NaN([par.nsim, 1, nsim_hor2]);
        timestep    = 1/par.stepyr;     
        r0(:,1,1)   = par.V_r0;
        mZ  = randn(par.nsim, nsim_hor2);
        for t = 2:nsim_hor2 %  eq (25-26), steeds met stappen van timestep
            % https://math.stackexchange.com/questions/1287634/implementing-ornstein-uhlenbeck-in-matlab:
            % tijdstap maakt uit voor asymptotische waardes van r0
            r0(:,1,t) = r0(:,1,t-1) + par.V_alp * (par.V_r0 - r0(:,1,t-1)) * timestep + par.V_rho * sqrt(timestep) * mZ(:,t-1);
        end
        % eq (28): definieer constante 'ultimate (forward) rente'
        Rinf = par.V_r0 - par.V_lam*(par.V_rho/par.V_alp) - 0.5*(par.V_rho/par.V_alp)^2;
        % dim 1: scenario's, dim 2: looptijd tellend vanaf 0, dim 3: simulatiehorizon
        temp_mat = repmat( permute( 1 - exp(-par.V_alp*repmat((0:timestep:maxmaturity)',1,nsim_hor2)), [3 1 2]), [par.nsim 1 1] );
        economy0.RTS_maturities   = linspace(0,maxmaturity,1+par.stepyr*maxmaturity);
        economy0.bond_portf_maturities   = vmatur;
        economy0.RTS = NaN(par.nsim, 1+maxmaturity/timestep, nsim_hor2);
        % eq (4) met exp(.):
        economy0.RTS(:,1,:)       = exp(r0);
        % eq (29) met exp():
        economy0.RTS(:,2:end,:)   = exp( Rinf + (r0-Rinf) .* temp_mat(:,2:end,:)./(par.V_alp*repmat(timestep:timestep:maxmaturity,[par.nsim 1 nsim_hor2])) ...
            + temp_mat(:,2:end,:).^2 .* par.V_rho.^2 ./ (4*par.V_alp^3.*repmat(timestep:timestep:maxmaturity,[par.nsim 1 nsim_hor2])) );
        economy0.euribor_return   = squeeze( economy0.RTS(:,economy0.RTS_maturities==1/par.stepyr,:).^(1/par.stepyr) ); % ones(par.nsim, nsim_hor1);
        RTSlocs                          = find(ismember(economy0.RTS_maturities,economy0.bond_portf_maturities(2:end)));
        % anders matchet het niet meer met de annuiteiten waardoor
        %  de renteafdekking niet meer exact klopt.
        economy0.bond_returns            = NaN(par.nsim, length(economy0.bond_portf_maturities), nsim_hor2-1);
        % looptijd nul is altijd continu tegen euribortarief:
        economy0.bond_returns(:,economy0.bond_portf_maturities==0,:)     = economy0.RTS(1:par.nsim, 1, 1:nsim_hor2-1);
        % langere looptijden worden iedere periode geherbalanceerd
        economy0.bond_returns(:,2:end,:) = economy0.RTS(:,RTSlocs,1:end-1) ...
                                                .^repmat(economy0.bond_portf_maturities(2:end),[par.nsim 1 nsim_hor2-1]) ...
                                               ./ economy0.RTS(:,RTSlocs-1,2:end) ...
                                                .^repmat(economy0.bond_portf_maturities(2:end)-1/par.stepyr,[par.nsim 1 nsim_hor2-1]);
    else
        if par.set_RTS > 1
            par.set_RTS = par.set_RTS - 1;
        end
        economy0.RTS                     = repmat(1+par.set_RTS,[par.nsim 1+par.stepyr*maxmaturity nsim_hor2]);       
        economy0.RTS_maturities          = linspace(0,maxmaturity, 1 + maxmaturity*par.stepyr);
        economy0.bond_portf_maturities   = vmatur;
        economy0.bond_returns            = NaN(par.nsim, length(economy0.bond_portf_maturities), nsim_hor2-1);
        economy0.bond_returns            = repmat( (1+par.set_RTS).^(1/par.stepyr),[par.nsim length(economy0.bond_portf_maturities) nsim_hor2-1]);
        economy0.euribor_return          = (1+par.set_RTS).^(1/par.stepyr) * all_ones1;
    end
    economy0.stock_returns = NaN(par.nsim, nsim_hor2);
    if strcmp(par.set_stocks,'scenarioset')
        economy0.stock_returns(:,2:end) = 1+economy.stock_return(1:par.nsim, 1:par.nsim_hor);
    else
        if strcmp(par.set_stocks,'Black Scholes')
            % simulate gross return using Black-Scholes model
            % log( S_t / S_{t-1} ) =  N( r + L*sig - sig^2/2, sig^2 );
            %      S_t / S_{t-1}   = LN( r + L*sig - sig^2/2, sig^2 );
            %   E[ S_t / S_{t-1} ] = exp(r + L*sig );
            % Var[ S_t / S_{t-1} ] = [ exp( sig^2 ) - 1 ] * exp( 2*(r + L*sig) );
            % drift = r + L*sig
            timestep                        = 1/par.stepyr;
            mu                              = log(economy0.euribor_return) + par.BS_risk_par * par.BS_vol_par * timestep;
            mZ2                             = randn(par.nsim, nsim_hor2);
            economy0.stock_returns(:,2:end) = exp( (mu(:,2:end) - .5 * timestep * par.BS_vol_par^2) + par.BS_vol_par *sqrt(timestep) * mZ2(:,2:end) );
        else
            if par.set_stocks > 1
                par.set_stocks = par.set_stocks - 1;
            end
            economy0.stock_returns(:,2:end) = (1 + par.set_stocks).^(1/par.stepyr);
        end
    end
    economy0.price_inflation = NaN(par.nsim, nsim_hor2);
    if strcmp(par.set_price_inflation,'scenarioset')
        economy0.price_inflation(:,2:end) = 1+economy.price_inflation(1:par.nsim, 1:par.nsim_hor);
    else
        economy0.price_inflation = (1+par.set_price_inflation).^(1/par.stepyr); % * ones(par.nsim, nsim_hor1);
    end
    economy0.wage_inflation = NaN(par.nsim, nsim_hor2);
    if strcmp(par.set_wage_inflation,'scenarioset')
        economy0.wage_inflation(:,2:end)  = 1+economy.wage_inflation(1:par.nsim, 1:par.nsim_hor);
    else
        if par.set_wage_inflation > 1
            par.set_wage_inflation = par.set_wage_inflation - 1;
        end
        economy0.wage_inflation = (1+par.set_wage_inflation).^(1/par.stepyr); % * ones(par.nsim, par.nsim_hor);
    end

end

function [RTS, maturities] = constructRTS(economy,stepyr)
    if ~exist('stepyr','var')
        stepyr = 1;
    end
    if stepyr>1
        economy.nominal_term_structure.a = from_yearly_to_periodically(economy.nominal_term_structure.a,stepyr,1);
        economy.nominal_term_structure.b = from_yearly_to_periodically(economy.nominal_term_structure.b,stepyr,1);
    end
    % construeer RTS met gesimuleeerde states
    RTS = NaN(economy.number_of_scenarios, length(economy.nominal_term_structure.maturities), economy.number_of_years+1);
    for s=1:economy.number_of_scenarios
        RTS(s,:,:) = exp( repmat( economy.nominal_term_structure.a(:), [1 economy.number_of_years+1]) ...
                    + economy.nominal_term_structure.b*squeeze(economy.state_variables_term_structure(s,:,:))) - 1;
    end
    maturities = economy.nominal_term_structure.maturities;
end

function M = from_yearly_to_periodically(v,n,dim)
    if ~exist('dim','var')
        dim = 1;
    end
    if dim==2
        v = v';
    end
    M = NaN(size(v).*[n 1]);
    stepsizes  = diff(v,1,dim)/n;
    for i=1:size(v,2)
        vM     = v(:,i)' + (-(n-1):0)'*[stepsizes(1) stepsizes(:,i)'];
        M(:,i) = vM(:);
    end
    if dim==2
        M = M';
    end
end

function [par, profiel,sheets_cell] = readInput(inputfile,sheet,sheets,bclearsheet)
    % bclearsheet:   true:  kolommen moeten hoe dan ook schoon
    %                false: kolommen hoeven niet per se schoon 
    % lees parameters in
    if isempty(sheet)
        [~,sheets_cell] = xlsfinfo(inputfile);
    else
        sheets_cell = {sheet};
    end
    % n sheet met leeftijdsprofielen, andere bevatten contracten
    numcells = length(sheets_cell); %-sum(ismember({'stats' 'profiel'}, sheets_cell));
    par      = cell(1, numcells);
    profiel  = cell(1,numcells);
    % lees leeftijdsprofielen in
    try
        [~,~,C]  = xlsread(inputfile,'profiel');
        [row,colvar] = find(strcmp(C,'age'));
        for i = colvar:size(C,2)
            s    = C{row,i};
            dat1 = cell2mat(C(row+1:size(C,1),i));
            try
                profiel0.(s) = dat1;
            catch % stop bij lege cel
                break
            end
        end
    catch
    end
    
    colLetters = xlsColNum2Str(1:500); % converteer kolomnummer naar kolomletter  
    for i=1:numcells
        if strcmp(sheets_cell{i},'main') ||strcmp(sheets_cell{i},'profiel') || strcmp(sheets_cell{i},'stats')

        else
            [~,~,C]       = xlsread(inputfile, sheets_cell{i});
            [~,colremove] = find(strcmp(C,'geomean'));
            if ~isempty(colremove)
                if colremove(1) >= 5
                    [C{:,colremove(1):end}] = deal(NaN);
                    xlswrite_wrap(inputfile, C(:,colremove(1):end), sheets_cell{i},colremove(1),1,colLetters,bclearsheet);
                end
            end
            [row,colvar]    = find(strcmp(C,'variable name'));
            [~,colmod]      = find(strcmp(C,'in model'));
            for j = row+1:size(C,1)
                s = C{j,colvar(1)};
                dat1 = C{j,colmod(1)};
                try
                    tmp = str2double(dat1);
                    if (~isempty(tmp) && ~isnan(tmp)) || strcmp(dat1,'NaN')
                        par{i}.(s) = tmp;
                    else
                        par{i}.(s) = dat1;
                    end
                catch % sla lege rijen over
                end
            end
            switch par{i}.scenarioset
                case {'componenten','all 1'}
                    par{i}.sfile       = [];
                otherwise
                    par{i}.sfile       = ['KNW_' par{i}.scenarioset '_P.mat'];
            end

            if all(~isnan(par{i}.asset_mix)) && ismember(par{i}.asset_mix, sheets)
                [~,~,Ctmp]  = xlsread(inputfile,'stats');
                Irow0       = strcmp(Ctmp(:,1), 'sheetnaam' );
                if (isempty(Irow0) || all(~Irow0)), Irow0 = 2; end
                Irow_stock  = strcmp(Ctmp(:,1), 'stocks_assets ' );
                Irow2       = strcmp(Ctmp(:,1), 'bonds_assets  ' );
                Irow3       = strcmp(Ctmp(:,1), 'euribor_assets' );
                Icol        = strcmp(Ctmp(Irow0 ,:), par{i}.asset_mix );
                par{i}.stock_sol   = cell2mat(Ctmp(Irow_stock,find(Icol,1)));
                par{i}.bond_sol    = cell2mat(Ctmp(Irow2,find(Icol,1)));
                par{i}.euribor_sol = cell2mat(Ctmp(Irow3,find(Icol,1)));
                
                num_ages          = par{i}.max_age + 1 - par{i}.first_exposure_age;
                profiel{i}.stocks = NaN(num_ages, 1);
                profiel{i}.stocks(:)  = cell2mat(Ctmp(Irow_stock,find(Icol,1)));
                % aanname: alleen in aandelen en n-jaarsobligaties
                profiel{i}.bond(:)    = cell2mat(Ctmp(Irow2,find(Icol,1)));
                profiel{i}.euribor(:) = cell2mat(Ctmp(Irow3,find(Icol,1)));
                
                clear Ctmp Irow0 Irow1 Irow2 Irow3
            end
            if strcmp(par{i}.asset_mix, 'lineair')
                % lineair profiel obv stockfracties op aantal leeftijden
                v_age    = [par{i}.first_exposure_age   par{i}.pension_age-1   ...
                    par{i}.pension_age   min(85,par{i}.max_age)   par{i}.max_age];
                v_stocks = [par{i}.stock_first_exposure_age  par{i}.stock_last_work_age  par{i}.stock_pension_age ...
                                par{i}.stock_85               par{i}.stock_max_age];
                I        = ~isnan(v_stocks);
                num_ages = par{i}.max_age + 1 - par{i}.first_exposure_age;
                profiel{i}.stocks = NaN(num_ages, 1);
                profiel{i}.stocks(1:(max(v_age(I))+1-v_age(1))) = linprofiles( v_age(1), v_age(I), v_stocks(I), 1); % par{i}.stepyr
                % aanname: alleen in aandelen en n-jaarsobligaties
                profiel{i}.bond    = 1 - profiel{i}.stocks;
                profiel{i}.euribor = zeros(size(profiel{i}.bond));
            end
            if strcmp(par{i}.linear_annuity_buy, 'lineair') || strcmp(par{i}.linear_annuity_buy, 'reciproke lineair')
                % (reciproke) lineair profiel obv stockfracties op aantal leeftijden
                if isnan(par{i}.annuity_fraction_first_buy_age) % neem lineaire stijging aan
                    v_age           = [par{i}.annuity_first_buy_age-1   par{i}.pension_age-1/par{i}.stepyr  par{i}.max_age];
                    v_frac_buy      = [0 1 1];
                else
                    v_age           = [par{i}.annuity_first_buy_age   par{i}.pension_age-1/par{i}.stepyr   par{i}.max_age];
                    v_frac_buy      = [par{i}.annuity_fraction_first_buy_age  1  1];
                end
                I                   = ~isnan(v_frac_buy);
                profiel{i}.annuity_buy = linprofiles(par{i}.first_exposure_age, v_age(I), v_frac_buy(I), 1, strcmp(par{i}.linear_annuity_buy, 'reciproke lineair'));
            end

            if exist('profiel0','var')
                profiel{i}     = profiel0;
                par{i}.max_age = min(par{i}.max_age, profiel{i}.age(1+find(profiel{i}.Psurv>0,1,'last')));
            else
                if strcmp(par{i}.asset_mix,'specifiek leeftijdsprofiel')
                    error('ontbrekend leeftijdsprofiel')
                end
            end
            
            if ~isempty(profiel{i})
                for s=fieldnames(profiel{i})'
                    v = profiel{i}.(s{1});
                    v(par{i}.max_age+1:end) = [];
                    profiel{i}.(s{1}) = v;
                end
                if ~isfield(profiel{i},'wage')
                    profiel{i}.wage      = ones(par{i}.max_age+1-par{i}.first_exposure_age,1);
                    profiel{i}.wage(1:par{i}.first_work_age-par{i}.first_exposure_age) = 0;
                end
                if ~isfield(profiel{i},'franchise')
                    profiel{i}.franchise = zeros(par{i}.max_age+1-par{i}.first_exposure_age,1);
                end
                if ~isfield(profiel{i},'Psurv')
                    profiel{i}.Psurv     = ones(par{i}.max_age+1-par{i}.first_exposure_age,1);
                    profiel{i}.Psurv(end)= 0;
                end
                if isfield(profiel{i},'age')
                    par{i}.first_exposure_age = max(par{i}.first_exposure_age, ...
                        profiel{i}.age(find(profiel{i}.wage>0,1)));
                end
            else
                profiel{i}.wage      = ones(par{i}.max_age+1-par{i}.first_exposure_age,1);
                profiel{i}.wage(1:par{i}.first_work_age-par{i}.first_exposure_age) = 0;
                profiel{i}.franchise = zeros(par{i}.max_age+1-par{i}.first_exposure_age,1);
                profiel{i}.Psurv     = ones(par{i}.max_age+1-par{i}.first_exposure_age,1);
                profiel{i}.Psurv(end)= 0;
            end
            if par{i}.stepyr~=1
                % maak lineaire interpolatie
                for s=fieldnames(profiel{i})'
                     profiel{i}.(s{1}) = from_yearly_to_periodically(profiel{i}.(s{1}),par{i}.stepyr,1);
                end
            end
        end
    end
end

function vv = rescale(vv, assets, Inonneg_ages)
    % herschaal zodanig dat negatieve totaalrendementen niet mogelijk zijn.
    % met andere woorden, de totale exposure is maximaal 1;
    if ~isempty(Inonneg_ages)
        v_tmp   = sum( vv(:,Inonneg_ages) .* assets(:,Inonneg_ages), 2) ./ sum(assets(:,Inonneg_ages), 2);
        vv(:,Inonneg_ages)  = vv(:,Inonneg_ages) ./ max(v_tmp,1);
    end
end

function v_intp = linprofiles(age0, v_age, p, stepyr, breciproke)
%LINPROFILES maak lineair profiel van interpolatieprofielen
% age0  is de eerste leeftijd in return vector v
% v_age is een vector met leeftijden
% p     is een vector met fracties
% breciproke is als reciproke lineair is: 1/n, 1/(n-1), 1/(n-2), ..., 1/1
% v_intp is de outputvector met lineair genterpoleerde fracties

    if ~exist('breciproke','var')
        breciproke = false;
    end
    v_age0 = v_age - age0 + 1;
    v_intp = zeros( v_age0(end)*stepyr, 1);
    for i=2:length(v_age0)
        if ~breciproke
            v_intp( v_age0(i-1)*stepyr:v_age0(i)*stepyr ) = linspace(p(i-1), p(i), stepyr*(v_age0(i)-v_age0(i-1))+1);
        else
            v_intp( v_age0(i-1)*stepyr:v_age0(i)*stepyr ) = 1./linspace(1/p(i-1), 1/p(i), stepyr*(v_age0(i)-v_age0(i-1))+1);
        end
    end
end

function [lastcol,lastrow] = xlswrite_wrap(file, data, sheet, firstcol, firstrow, colLetters, bclearsheet)
% write data at range of sheet on file
% wrapper function of xlswrite
% firstcol   is first column number
% firstrow   is first row number
% colLetters is an optional cell array with column letters in each cell
% bclearsheet:   true:  kolommen moeten hoe dan ook schoon
%                false: kolommen hoeven niet per se schoon 
% lastcol    is rightmost column number
    if ~exist('bclearsheet','var')
        bclearsheet = true;
    end
    range = [colLetters{firstcol} num2str(firstrow)];
    try
        xlswrite(file, data, sheet, range);
    catch
        if bclearsheet % sheet moet hoe dan ook schoon
            pause(1); % soms is wachten op voorgaand schrijven nodig
            try
                xlswrite(file, data, sheet, range);
            catch
                pause(1); % soms is wachten op voorgaand schrijven nodig
                try
                    xlswrite(file, data, sheet, range);
                catch
                    disp(['Please close file ' file])
                    disp('Press any key to continue')
                    pause;
                    xlswrite(file, data, sheet, range);
                end
            end
        end
    end
    lastcol = firstcol + size(data,2) - 1;
    lastrow = firstrow + size(data,1) - 1;
end

function [colChar]=xlsColNum2Str(colNum)
%XLSCOLNUM2STR takes in an array of numbers and returns a cellular array
%of the same size with cell of corresponding Excel column names.
%
%For example:
%n=[1  10;
%   53 256]
%c=xlsColNum2Str(n);
%c={'A' , 'J';
%   'BA', 'IV'}
%Note: up to Excel 2003 the number of columns was limited to 256, as of
%Excel 2007 the number of columns has increased to 16,384 or 'XFD'
%This function is designed to take accept any integer so proper handling 
%of the number of columns should be taken care of outside this function
% https://nl.mathworks.com/matlabcentral/fileexchange/15748-excel-column-number-to-column-name?requestedDomain=true
    colChar=cell(size(colNum)); %blank cell array
    
    % find max number of characters (AA n=2)
    numOfChars=ceil(max(colNum)/26)-1;
    n=1;
    
    while numOfChars>=1
        numOfChars=ceil(numOfChars/26)-1;
        n=n+1;
    end    
    
    remainder=num2cell(colNum);
    
    for s=n:-1:1
        if s>1
            %find limits
            % if n=2 then the columns go from AA to ZZ or 27 to 702
            L=sum(26.^(1:s-1))+1; % lower limit
            U=sum(26.^(1:s));   %upper limit
            %place current character to right of previous
            colChar(colNum>=L & colNum<=U) = ...
                cellfun(@(x,y) ([x char(ceil((y-(L-1))/26^(s-1))+64)]),...
                colChar(colNum>=L & colNum<=U),...      % x
                remainder(colNum>=L & colNum<=U),...    % y
                'UniformOutput',false);
            %calculate the remaining string
            %for example if last string was 'ABA' the 'A' was placed to the
            %right of the previous string and now 'BA' is remaining
            remainder(colNum>=L & colNum<=U)=...
                cellfun(@(x,y) (y-26^(s-1)*(double(x(end))-64)),...
                colChar(colNum>=L & colNum<=U),...
                remainder(colNum>=L & colNum<=U),'UniformOutput',false);
            colNum=cell2mat(remainder);
        else
             colChar=cellfun(@(x,y) ([x char(y+64)]),...
                 colChar,remainder,'UniformOutput',false);
        end
    end    
end

function Acoh = extractDiag(A,dim)
% deze functie kan eventueel cohorteffecten bepalen
    % A:      three dimensional matrix
    % dim(1): scenario
    % dim(2): age
    % dim(3): time
    if ~exist('dim','var') || isempty(dim)
        dim = 1:3; % default: dim=1:3
    end
    [nsim,numages,nhor] = size(A);
    A           = permute(A, dim(3:-1:1));
    firstelem   = [nhor:-1:1 nhor*(1:numages-1)+1];
    firstperiod = [nhor:-1:1 ones(1,numages-1)];
    lastperiod  = min(firstperiod+numages-1, ...
                    [nhor*ones(1,length(firstperiod)-min(nhor,numages)) min(nhor,numages):-1:1] );
    Acoh        = NaN(nhor, length(firstelem), nsim);
    coh         = 0;
    stepsize    = nhor+1;
    % Acoh
    % first dim: time
    % second dim: cohort
    % third dim: scenario
    for i = 1:length(firstelem)
        coh = coh+1;
        numperiods = lastperiod(i)-firstperiod(i)+1;
        Acoh(firstperiod(i):lastperiod(i),coh,:) = ...
            A( repmat((firstelem(i)+stepsize*(0:numperiods-1))',[1 nsim] ) ...
                + nhor*numages*repmat(0:(nsim-1),[numperiods 1]) );
    end
    Acoh = Acoh(:,end:-1:1,:); % rank cohorts from oldest to youngest
    [~,dimloc] = sort(dim(3:-1:1));
    Acoh       = permute(Acoh, dimloc);
end